import json
from typing import cast
import tornado
from tinybird.integrations.vercel import VercelIntegration, VercelIntegrationException, VercelIntegrationPhase, VercelIntegrationService
from tinybird.token_scope import scopes

from tinybird.user import User, UserAccount, UserAccounts
from .views.base_test import BaseTest, TBApiProxyAsync, TBApiProxy

from unittest.mock import patch


class TestVercelIntegrationService(BaseTest):
    def setUp(self):
        super(TestVercelIntegrationService, self).setUp()
        self.tb_api_proxy_async: TBApiProxyAsync = TBApiProxyAsync(self)
        self.tb_api_proxy: TBApiProxy = TBApiProxy(self)
        self.user_account: UserAccount = UserAccount.get_by_id(self.USER_ID)

    async def reset_user(self) -> None:
        integrations = self.user_account.get_integrations()
        if not len(integrations):
            return
        integration = integrations[0]
        self.user_account = await VercelIntegrationService.remove_integration(self.user_account,
                                                                              integration.id,
                                                                              remove_remote=False)

    async def _create_integration(self) -> VercelIntegration:
        integration = await VercelIntegrationService.get_integration_for_user(self.user_account, 'dummy_access_code')
        assert integration.access_code
        assert integration.integration_phase == VercelIntegrationPhase.INSTALLING

        self.user_account = await UserAccounts.add_integration(self.user_account, 'vercel', integration.id)
        assert len(self.user_account.get_integration_info_by_type('vercel')) == 1
        return integration

    async def _finalize_install(self, integration: VercelIntegration) -> VercelIntegration:
        api_mock: str = json.dumps({
            "token_type": "Bearer",
            "access_token": "xEbuzM1ZAJ46afITQlYqH605",
            "installation_id": "icfg_ijgU3GzRxyz7nGpwwsSmbNkI",
            "user_id": "2tUzTmv4ljvFVHifEf9TGdpH",
            "team_id": None
        })

        with patch.object(VercelIntegrationService, '_vercel_post', return_value=(200, api_mock)):
            integration = await VercelIntegrationService.finalize_install(integration)
            assert not integration.access_code
            assert integration.integration_phase == VercelIntegrationPhase.CONFIGURED
            return integration

    @tornado.testing.gen_test
    async def test_create(self) -> None:
        await self.reset_user()
        _ = await self._create_integration()

    @tornado.testing.gen_test
    async def test_create_and_finalize(self) -> None:
        await self.reset_user()
        integration = await self._create_integration()
        installed = await self._finalize_install(integration)
        assert installed.id == integration.id

    @tornado.testing.gen_test
    async def test_create_and_finalize_twice(self) -> None:
        await self.reset_user()
        integration = await self._create_integration()
        installed = await self._finalize_install(integration)
        assert not installed.access_code
        assert installed.integration_phase == VercelIntegrationPhase.CONFIGURED

    @tornado.testing.gen_test
    async def test_create_and_fetch(self) -> None:
        await self.reset_user()
        integration = await self._create_integration()
        dummy = await VercelIntegrationService.get_integration_for_user(self.user_account)
        assert dummy.id == integration.id

    @tornado.testing.gen_test
    async def test_add_binding(self) -> None:
        await self.reset_user()
        integration = await self._create_integration()
        integration = await self._finalize_install(integration)

        workspace = User.get_by_id(self.WORKSPACE_ID)
        tk: str = cast(str, workspace.get_token_for_scope(scopes.ADMIN))

        with patch.object(VercelIntegrationService, '_vercel_post', return_value=(200, '{}')):
            integration = await VercelIntegrationService.add_bindings(integration, 'test_project', self.WORKSPACE_ID, [tk])
            assert len(integration.get_bindings()) == 1

        with patch.object(VercelIntegrationService, '_vercel_post', return_value=(200, '{}')):
            try:
                integration = await VercelIntegrationService.add_bindings(integration, 'test_project', self.WORKSPACE_ID, ['unknown_token'])
            except VercelIntegrationException:
                pass
            assert len(integration.get_bindings()) == 1



import unittest
import uuid

from tinybird.hook import Hook
from tinybird.user import User, UserAccount


class BaseTestHook(Hook):
    def __init__(self, user):
        super().__init__(user)
        self.called_before_create = False
        self.called_after_create = False
        self.called_tear_down = False
        self.called_on_error = False
        self.captured_error = None

    def on_error(self, datasource, error):
        self.called_on_error = True
        self.captured_error = error

    def tear_down(self, datasource):
        self.called_tear_down = True


class HappyCaseHook(BaseTestHook):

    def before_create(self, datasource):
        self.called_before_create = True

    def after_create(self, datasource):
        self.called_after_create = True
        self.tear_down(datasource)


class AfterCreateFailsHook(BaseTestHook):

    def before_create(self, datasource):
        self.called_before_create = True

    def after_create(self, datasource):
        self.called_after_create = True
        raise RuntimeError('Failed before tear_down')


class TestHooks(unittest.TestCase):

    def setUp(self):
        self.WORKSPACE = f'test_hooks_{uuid.uuid4().hex}'
        self.USER = f'{self.WORKSPACE}@example.com'

        self.user = UserAccount.register(self.USER, 'pass')
        self.workspace = User.register(
            name=self.WORKSPACE,
            admin=self.user.id
        )

    def tearDown(self):
        User._delete(self.workspace.id)
        user_account = UserAccount.get_by_email(self.USER)
        UserAccount._delete(user_account.id)

    def test_happy_case(self):
        ds = self.workspace.add_datasource('fake_datasource')

        hook = HappyCaseHook(self.workspace)
        ds.install_hook(hook)

        try:
            for h in ds.hooks:
                h.before_create(ds)

            for h in ds.hooks:
                h.after_create(ds)
        except Exception:
            for h in ds.hooks:
                h.on_error(ds)

        self.assertTrue(hook.called_before_create)
        self.assertTrue(hook.called_after_create)
        self.assertTrue(hook.called_tear_down)
        self.assertFalse(hook.called_on_error)

    def test_after_create_fails_but_tear_down_still_called(self):
        ds = self.workspace.add_datasource('fake_datasource')

        hook = AfterCreateFailsHook(self.workspace)
        ds.install_hook(hook)

        try:
            for h in ds.hooks:
                h.before_create(ds)

            for h in ds.hooks:
                h.after_create(ds)
        except Exception as e:
            self.assertEqual(hook.captured_error, e)

        self.assertTrue(hook.called_before_create)
        self.assertTrue(hook.called_after_create)
        self.assertTrue(hook.called_tear_down)
        self.assertTrue(hook.called_on_error)

    def test_after_create_two_hooks_fails_first(self):
        ds = self.workspace.add_datasource('fake_datasource')

        fails_hook = AfterCreateFailsHook(self.workspace)
        ds.install_hook(fails_hook)
        happy_hook = HappyCaseHook(self.workspace)
        ds.install_hook(happy_hook)

        try:
            for h in ds.hooks:
                h.before_create(ds)

            for h in ds.hooks:
                h.after_create(ds)
        except Exception as e:
            self.assertEqual(fails_hook.captured_error, e)
            for h in ds.hooks:
                h.on_error(ds, e)

        self.assertTrue(fails_hook.called_before_create)
        self.assertTrue(fails_hook.called_after_create)
        self.assertTrue(fails_hook.called_tear_down)
        self.assertTrue(fails_hook.called_on_error)

        self.assertTrue(happy_hook.called_before_create)
        self.assertFalse(happy_hook.called_after_create)
        self.assertTrue(happy_hook.called_tear_down)
        self.assertTrue(happy_hook.called_on_error)



import uuid
import tornado
from urllib.parse import urlencode
from unittest.mock import patch
import json


from tinybird.user import UserAccount, Users, scopes, User, UserWorkspaceRelationship
from tinybird.constants import CHCluster
from tinybird.workspace_service import WorkspaceService
from .views.base_test import BaseTest, TBApiProxyAsync, TBApiProxy
from .utils import exec_sql
from tinybird.ch import ch_table_details_async, ch_table_schema_async
from tinybird.data_connector import DataConnector
from tinybird.datasource import KafkaBranchDatasource


class TestWorkspaceService(BaseTest):

    def setUp(self):
        super(TestWorkspaceService, self).setUp()
        self.tb_api_proxy_async = TBApiProxyAsync(self)
        self.tb_api_proxy = TBApiProxy(self)
        self.user_account = UserAccount.get_by_id(self.USER_ID)

        # create origin workspace using cluster
        self.cluster = CHCluster('tinybird', User.default_database_server)
        workspace_name = f"ws_to_clone_{uuid.uuid4().hex}"
        self.workspace = self.tb_api_proxy.register_workspace(workspace_name, self.user_account, self.cluster)
        self.workspace_token = Users.get_token_for_scope(self.workspace, scopes.ADMIN)
        self.user_token = UserAccount.get_token_for_scope(self.user_account, scopes.AUTH)

    def assertEqualDatasource(self, datasource_a, datasource_b):
        attrs = ['id', 'name', '_description', 'tags', 'json_deserialization']
        for attr in attrs:
            self.assertEqual(getattr(datasource_a, attr), getattr(datasource_b, attr), attr)

    def assertEqualPipe(self, pipe_a, pipe_b):
        attrs = ['id', 'name', 'description', 'endpoint']
        for attr in attrs:
            self.assertEqual(getattr(pipe_a, attr), getattr(pipe_b, attr), attr)
        self.assertEqual(len(pipe_a.pipeline.nodes), len(pipe_b.pipeline.nodes))
        self.assertEqual(pipe_a.pipeline.to_json(), pipe_b.pipeline.to_json())

    def assertEqualToken(self, token_a, token_b):
        attrs = ['name', 'scopes', 'origin']
        for attr in attrs:
            self.assertEqual(token_a.to_dict()[attr], token_b.to_dict()[attr])

    async def _test_clone_datasources(self):
        # create new workspace to clone datasource
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        new_workspace_token = Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        # clone datasource in new workspace
        await WorkspaceService.clone_datasources(self.workspace, new_workspace)
        new_workspace = User.get_by_id(new_workspace.id)
        self.assertEqual(len(self.workspace.get_datasources()), len(new_workspace.get_datasources()))
        cloned_datasource = new_workspace.get_datasource(self.datasource.id)
        self.assertEqualDatasource(self.datasource, cloned_datasource)
        table_details_a = await ch_table_details_async(self.datasource.id, self.workspace.database_server,
                                                       database=self.workspace.database)
        table_details_b = await ch_table_details_async(self.datasource.id, new_workspace.database_server,
                                                       database=new_workspace.database)
        self.assertEqual(table_details_a.to_json(), table_details_b.to_json())
        schema_a = await ch_table_schema_async(self.datasource.id, self.workspace.database_server,
                                               self.workspace.database)
        schema_b = await ch_table_schema_async(self.datasource.id, new_workspace.database_server,
                                               new_workspace.database)
        self.assertEqual(schema_a, schema_b)

        # check quarantine is available
        self.assertEqual(0, await self._query(new_workspace_token, f"select count() from {self.datasource.name}_quarantine"))

        # ds_ops_log
        self.expect_ops_log({'event_type': 'create',
                             'datasource_name': self.datasource.name,
                             'options': {'source': 'branch'}}, workspace=new_workspace)

    @tornado.testing.gen_test
    async def test_clone_datasources(self):
        # create datasource
        ds_name = f"ds_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            self.workspace_token,
            ds_name,
            schema
        )
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        await self._test_clone_datasources()

    @tornado.testing.gen_test
    async def test_clone_datasources_ndjson(self):
        # create datasource
        ds_name = f"ds_to_clone_ndjson_{uuid.uuid4().hex}"
        schema = "id String `json:$.id`"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            self.workspace_token,
            ds_name,
            schema,
            format='ndjson',
        )
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        await self._test_clone_datasources()

    @tornado.testing.gen_test
    async def test_clone_shared_datasources(self):
        # create new workspace to clone datasource
        sharing_workspace_name = f"ws_sharing_{uuid.uuid4().hex}"
        new_sharing_workspace = await self.tb_api_proxy_async.register_workspace(sharing_workspace_name, self.user_account, self.cluster)
        new_shared_workspace_token = Users.get_token_for_scope(new_sharing_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_sharing_workspace)

        # create datasource
        ds_name = f"ds_shared_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            new_shared_workspace_token,
            ds_name,
            schema
        )
        new_sharing_workspace = User.get_by_id(new_sharing_workspace.id)
        self.datasource = new_sharing_workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(new_sharing_workspace, self.datasource)

        # share Data Sources between them
        await self.tb_api_proxy_async.share_datasource_with_another_workspace(
            token=self.user_token,
            datasource_id=self.datasource.id,
            origin_workspace_id=new_sharing_workspace.id,
            destination_workspace_id=self.workspace.id,
            expect_notification=False
        )

        self.workspace = Users.get_by_id(self.workspace.id)
        self.assertEqual(len(self.workspace.get_datasources()), 1)
        self.datasource = self.workspace.get_datasources()[0]
        self.assertEqual(self.datasource.name, f"{new_sharing_workspace.name}.{ds_name}")

        # create new workspace to clone datasource
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account, self.cluster)
        new_workspace_token = Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        # set as branch manually
        new_workspace.origin = self.workspace.id
        new_workspace.save()

        # clone datasource in new workspace
        await WorkspaceService.clone_datasources(self.workspace, new_workspace)
        new_workspace = User.get_by_id(new_workspace.id)
        self.assertEqual(len(self.workspace.get_datasources()), len(new_workspace.get_datasources()))
        cloned_datasource = new_workspace.get_datasource(self.datasource.id, include_read_only=True)
        self.assertEqualDatasource(self.datasource, cloned_datasource)
        table_details_a = await ch_table_details_async(self.datasource.id, new_sharing_workspace.database_server, database=new_sharing_workspace.database)
        table_details_b = await ch_table_details_async(self.datasource.id, new_workspace.database_server, database=new_workspace.database)
        self.assertEqual(table_details_a.to_json(), table_details_b.to_json())
        schema_a = await ch_table_schema_async(self.datasource.id, new_sharing_workspace.database_server, new_sharing_workspace.database)
        schema_b = await ch_table_schema_async(self.datasource.id, new_workspace.database_server, new_workspace.database)
        self.assertEqual(schema_a, schema_b)

        # check ds and quarantine are available
        self.assertEqual(0, await self._query(new_workspace_token, f"select count() from {cloned_datasource.name}"))
        self.assertEqual(0, await self._query(new_workspace_token, f"select count() from {cloned_datasource.name}_quarantine"))

    @tornado.testing.gen_test
    @patch('tinybird.kafka_utils.KafkaUtils.get_kafka_topic_group', return_value={'response': 'ok'})
    async def test_clone_kafka_datasource(self, _kafka_utils_mock):
        params = {
            'token': self.workspace_token,
            'kafka_bootstrap_servers': 'localhost:9093',
            'kafka_security_protocol': 'plaintext',
            'kafka_sasl_plain_username': '',
            'kafka_sasl_plain_password': '',
            'name': 'kafka_1',
            'service': 'kafka',
            'kafka_sasl_mechanism': 'PLAIN'
        }

        url = f"/v0/connectors?{urlencode(params)}"
        response = await self.tb_api_proxy_async._fetch(url, method='POST', data='')
        self.assertEqual(response.status_code, 200)
        result = json.loads(response.content)

        ds_name = f"ds_kafka_{uuid.uuid4().hex}"
        data_connector = DataConnector.get_by_id(result['id'])
        params = {
            'token': self.workspace_token,
            'mode': 'create',
            'name': ds_name,
            'connector': data_connector.id,
            'kafka_topic': 'kafka_topic',
            'kafka_group_id': 'kafka_group_id'
        }

        create_url = f"/v0/datasources?{urlencode(params)}"
        response = await self.tb_api_proxy_async._fetch(create_url, method='POST', data='')
        ds_response = json.loads(response.content)
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        # create new workspace to clone datasource
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account, self.cluster)
        new_workspace_token = Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        # set as branch manually
        new_workspace.origin = self.workspace.id
        new_workspace.save()

        # clone datasource in new workspace
        await WorkspaceService.clone_datasources(self.workspace, new_workspace)
        new_workspace = User.get_by_id(new_workspace.id)
        self.assertEqual(len(self.workspace.get_datasources()), len(new_workspace.get_datasources()))
        cloned_datasource = new_workspace.get_datasource(self.datasource.id, include_read_only=True)
        self.assertEqualDatasource(self.datasource, cloned_datasource)
        self.assertTrue(isinstance(cloned_datasource, KafkaBranchDatasource))
        self.assertNotEqual(cloned_datasource.to_json()['type'], 'kafka')
        self.assertEqual(self.datasource.to_dict().get('connector'), cloned_datasource.origin_connector_id)
        table_details_a = await ch_table_details_async(self.datasource.id, self.workspace.database_server,
                                                       database=self.workspace.database)
        table_details_b = await ch_table_details_async(self.datasource.id, new_workspace.database_server,
                                                       database=new_workspace.database)
        self.assertEqual(table_details_a.to_json(), table_details_b.to_json())
        schema_a = await ch_table_schema_async(self.datasource.id, self.workspace.database_server,
                                               self.workspace.database)
        schema_b = await ch_table_schema_async(self.datasource.id, new_workspace.database_server,
                                               new_workspace.database)
        self.assertEqual(schema_a, schema_b)

        # check ds and quarantine are available
        self.assertEqual(0, await self._query(new_workspace_token, f"select count() from {cloned_datasource.name}"))
        self.assertEqual(0, await self._query(new_workspace_token,
                                              f"select count() from {cloned_datasource.name}_quarantine"))

    @tornado.testing.gen_test
    async def test_clone_pipe_endpoint(self):
        # create pipe in workspace origin
        pipe_name = f"pipe_to_clone_{uuid.uuid4().hex}"
        Users.add_pipe(self.workspace, self.pipe_name, f"select * from {self.datasource.name}")
        await self.tb_api_proxy_async.create_pipe_endpoint(self.workspace, self.workspace_token, pipe_name,
                                                           "select count() from numbers(1,10)")
        self.workspace = User.get_by_id(self.workspace.id)
        pipe = self.workspace.get_pipe(pipe_name)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account, self.cluster)
        new_workspace_token = Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        # clone pipe in new workspace
        await WorkspaceService.clone_pipes(self.workspace, new_workspace)
        new_workspace = User.get_by_id(new_workspace.id)
        cloned_pipe = new_workspace.get_pipe(pipe.id)

        self.assertEqualPipe(cloned_pipe, pipe)

        # Request to pipe endpoint in new workspace
        params = {
            'token': new_workspace_token
        }
        response = await self.fetch_async(f'/v0/pipes/{pipe_name}.json?{urlencode(params)}')
        self.assertEqual(response.code, 200)

    @tornado.testing.gen_test
    async def test_clone_pipe_materializes(self):
        # create datasource
        ds_name = f"ds_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            self.workspace_token,
            ds_name,
            schema
        )
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        # create a pipe's node with a view to materialize
        pipe_name = 'test_mat_view_to_clone'
        view_name = 'mat_view_node'
        target_ds_name = 'mat_view_node_ds'
        query = f'select q * 2 as b from {ds_name}'
        await self.tb_api_proxy_async.create_pipe_mv(self.workspace, self.workspace_token, pipe_name, view_name,
                                                     target_ds_name, query)
        self.workspace = User.get_by_id(self.workspace.id)
        pipe_mv = self.workspace.get_pipe(pipe_name)
        self.workspace.get_datasource(target_ds_name)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        # we need to clone datasources in new workspace first
        await WorkspaceService.clone_datasources(self.workspace, new_workspace)

        # clone pipe in new workspace
        new_workspace = User.get_by_id(new_workspace.id)
        await WorkspaceService.clone_pipes(self.workspace, new_workspace)
        new_workspace = User.get_by_id(new_workspace.id)
        cloned_pipe = new_workspace.get_pipe(pipe_mv.id)

        self.assertEqualPipe(cloned_pipe, pipe_mv)

        node_view_id = next(node.id for node in cloned_pipe.pipeline.nodes if node.materialized)

        query = f"""SELECT count() as c
                    FROM system.tables
                    WHERE
                        database = '{new_workspace.database}'
                        and name = '{node_view_id}'
                    FORMAT JSON"""
        r = exec_sql(new_workspace.database, query)
        table_exists = int(r['data'][0]['c']) == 1
        self.assertTrue(table_exists)

    @tornado.testing.gen_test
    async def test_clone_pipe_materializes_dependant_to_other_pipe(self):
        # create datasource
        ds_name = f"ds_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            self.workspace_token,
            ds_name,
            schema
        )
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        # MV dependant node
        pipe_name = f"pipe_dependant_{uuid.uuid4().hex}"
        await self.tb_api_proxy_async.create_pipe_endpoint(self.workspace, self.workspace_token, pipe_name,
                                                           f"select * from {self.datasource.name}")
        self.workspace = User.get_by_id(self.workspace.id)

        # create a pipe's node with a view to materialize
        pipe_mv_name = 'test_mat_view_to_clone'
        view_name = 'mat_view_node'
        target_ds_name = 'mat_view_node_ds'
        query = f'select q * 2 as c from {pipe_name}'
        await self.tb_api_proxy_async.create_pipe_mv(self.workspace, self.workspace_token, pipe_mv_name, view_name,
                                                     target_ds_name, query)
        self.workspace = User.get_by_id(self.workspace.id)
        pipe_mv = self.workspace.get_pipe(pipe_mv_name)
        self.workspace.get_datasource(target_ds_name)

        self.workspace = User.get_by_id(self.workspace.id)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        # we need to clone datasources in new workspace first
        await WorkspaceService.clone_datasources(self.workspace, new_workspace)

        # clone pipe in new workspace
        new_workspace = User.get_by_id(new_workspace.id)
        await WorkspaceService.clone_pipes(self.workspace, new_workspace)
        new_workspace = User.get_by_id(new_workspace.id)
        cloned_pipe = new_workspace.get_pipe(pipe_mv.id)

        self.assertEqualPipe(cloned_pipe, pipe_mv)

        node_view_id = next(node.id for node in cloned_pipe.pipeline.nodes if node.materialized)

        query = f"""SELECT count() as c
                    FROM system.tables
                    WHERE
                        database = '{new_workspace.database}'
                        and name = '{node_view_id}'
                    FORMAT JSON"""
        r = exec_sql(new_workspace.database, query)
        table_exists = int(r['data'][0]['c']) == 1
        self.assertTrue(table_exists)

    @tornado.testing.gen_test
    async def test_clone_tokens(self):
        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        await WorkspaceService.clone_tokens(self.workspace, new_workspace)
        new_workspace = User.get_by_id(new_workspace.id)

        self.assertEqual(len(self.workspace.get_tokens()), len(new_workspace.get_tokens()))
        for token in self.workspace.get_tokens():
            branch_token = new_workspace.get_token(token.name)
            self.assertEqualToken(token, branch_token)

    @tornado.testing.gen_test
    async def test_compare_workspaces_no_changes(self):
        # create datasource
        ds_name = f"ds_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            self.workspace_token,
            ds_name,
            schema
        )
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        # create pipe in workspace origin
        pipe_name = f"pipe_to_clone_{uuid.uuid4().hex}"
        await self.tb_api_proxy_async.create_pipe_endpoint(self.workspace, self.workspace_token, pipe_name,
                                                           "select count() from numbers(1,10)")
        self.workspace = User.get_by_id(self.workspace.id)
        self.workspace.get_pipe(pipe_name)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        await WorkspaceService.clone_datasources(self.workspace, new_workspace)
        await WorkspaceService.clone_pipes(self.workspace, new_workspace)

        self.workspace = User.get_by_id(self.workspace.id)
        new_workspace = User.get_by_id(new_workspace.id)

        workspace_diff = await WorkspaceService.compare_workspaces(self.workspace, new_workspace)
        self.assertEqual(workspace_diff.datasources, {'new': [], 'modified': [], 'deleted': [], 'only_metadata': []})
        self.assertEqual(workspace_diff.pipes, {'new': [], 'modified': [], 'deleted': [], 'only_metadata': []})

    @tornado.testing.gen_test
    async def test_compare_workspaces_deleted_resources(self):
        # create datasource
        ds_name = f"ds_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            self.workspace_token,
            ds_name,
            schema
        )
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        # create pipe in workspace origin
        pipe_name = f"pipe_to_clone_{uuid.uuid4().hex}"
        await self.tb_api_proxy_async.create_pipe_endpoint(self.workspace, self.workspace_token, pipe_name,
                                                           "select count() from numbers(1,10)")
        self.workspace = User.get_by_id(self.workspace.id)
        pipe = self.workspace.get_pipe(pipe_name)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        self.workspaces_to_delete.append(new_workspace)

        workspace_diff = await WorkspaceService.compare_workspaces(self.workspace, new_workspace)
        self.assertEqual(workspace_diff.datasources, {'new': [], 'modified': [], 'deleted': [{'id': self.datasource.id,
                                                                                              'name': self.datasource.name}],
                                                      'only_metadata': []})
        self.assertEqual(workspace_diff.pipes, {'new': [], 'modified': [], 'deleted': [{'id': pipe.id, 'name': pipe.name}], 'only_metadata': []})

    @tornado.testing.gen_test
    async def test_compare_workspaces_new_resources(self):
        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        self.workspaces_to_delete.append(new_workspace)

        new_workspace_token = Users.get_token_for_scope(new_workspace, scopes.ADMIN)

        # create datasource
        ds_name = f"ds_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            new_workspace_token,
            ds_name,
            schema
        )
        new_workspace = User.get_by_id(new_workspace.id)
        self.datasource = new_workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(new_workspace, self.datasource)

        # create pipe in workspace origin
        pipe_name = f"pipe_to_clone_{uuid.uuid4().hex}"
        await self.tb_api_proxy_async.create_pipe_endpoint(new_workspace, new_workspace_token, pipe_name,
                                                           "select count() from numbers(1,10)")
        self.workspace = User.get_by_id(self.workspace.id)
        new_workspace = User.get_by_id(new_workspace.id)
        pipe = new_workspace.get_pipe(pipe_name)

        workspace_diff = await WorkspaceService.compare_workspaces(self.workspace, new_workspace)
        self.assertEqual(workspace_diff.datasources, {'new': [{'id': self.datasource.id, 'name': self.datasource.name}], 'modified': [], 'deleted': [], 'only_metadata': []})
        self.assertEqual(workspace_diff.pipes, {'new': [{'id': pipe.id, 'name': pipe.name}], 'modified': [], 'deleted': [], 'only_metadata': []})

    @tornado.testing.gen_test
    async def test_compare_workspaces_modified_resources(self):
        # create datasource
        ds_name = f"ds_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            self.workspace_token,
            ds_name,
            schema
        )
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        # create pipe in workspace origin
        pipe_name = f"pipe_to_clone_{uuid.uuid4().hex}"
        await self.tb_api_proxy_async.create_pipe_endpoint(self.workspace, self.workspace_token, pipe_name,
                                                           "select count() from numbers(1,10)")
        self.workspace = User.get_by_id(self.workspace.id)
        self.workspace.get_pipe(pipe_name)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        await WorkspaceService.clone_datasources(self.workspace, new_workspace)
        await WorkspaceService.clone_pipes(self.workspace, new_workspace)

        self.workspace = User.get_by_id(self.workspace.id)
        new_workspace = User.get_by_id(new_workspace.id)

        # modify datasource
        sql = f"ALTER TABLE {new_workspace.database}.{self.datasource.id} ADD COLUMN added UInt32 FIRST;"
        exec_sql(new_workspace.database, sql)

        # modify pipe
        pipe = new_workspace.get_pipe(pipe_name)
        await Users.drop_endpoint_of_pipe_node_async(new_workspace.id, pipe.id, pipe.pipeline.last().id)

        new_workspace = User.get_by_id(new_workspace.id)

        workspace_diff = await WorkspaceService.compare_workspaces(self.workspace, new_workspace)
        self.assertEqual(workspace_diff.datasources, {'new': [], 'modified': [{'id': self.datasource.id, 'name': self.datasource.name}], 'deleted': [], 'only_metadata': []})
        self.assertEqual(workspace_diff.pipes, {'new': [], 'modified': [{'id': pipe.id, 'name': pipe.name}], 'deleted': [], 'only_metadata': []})

    @tornado.testing.gen_test
    async def test_compare_workspaces_only_metadata_modified_resources(self):
        # create datasource
        ds_name = f"ds_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            self.workspace_token,
            ds_name,
            schema
        )
        self.workspace = User.get_by_id(self.workspace.id)
        self.datasource = self.workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(self.workspace, self.datasource)

        # create pipe in workspace origin
        pipe_name = f"pipe_to_clone_{uuid.uuid4().hex}"
        await self.tb_api_proxy_async.create_pipe_endpoint(self.workspace, self.workspace_token, pipe_name,
                                                           "select count() from numbers(1,10)")
        self.workspace = User.get_by_id(self.workspace.id)
        self.workspace.get_pipe(pipe_name)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        await WorkspaceService.clone_datasources(self.workspace, new_workspace)
        await WorkspaceService.clone_pipes(self.workspace, new_workspace)

        self.workspace = User.get_by_id(self.workspace.id)
        new_workspace = User.get_by_id(new_workspace.id)

        pipe = new_workspace.get_pipe(pipe_name)
        await Users.alter_datasource_name(new_workspace, self.datasource.id, f"modified_name_{self.datasource.name}")
        await Users.alter_pipe_async(new_workspace, pipe.id, f"modified_name_{pipe_name}")
        new_workspace = User.get_by_id(new_workspace.id)

        workspace_diff = await WorkspaceService.compare_workspaces(self.workspace, new_workspace)
        self.assertEqual(workspace_diff.datasources, {'new': [], 'modified': [], 'deleted': [], 'only_metadata': [{'id': self.datasource.id, 'name': self.datasource.name}]})
        self.assertEqual(workspace_diff.pipes, {'new': [], 'modified': [], 'deleted': [], 'only_metadata': [{'id': pipe.id, 'name': pipe.name}]})

    @tornado.testing.gen_test
    async def test_compare_workspaces_no_changes_with_shared_datasource(self):
        sharing_workspace_name = f"ws_sharing_{uuid.uuid4().hex}"
        new_sharing_workspace = await self.tb_api_proxy_async.register_workspace(sharing_workspace_name,
                                                                                 self.user_account, self.cluster)
        new_shared_workspace_token = Users.get_token_for_scope(new_sharing_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_sharing_workspace)

        # create datasource
        ds_name = f"ds_shared_to_clone_{uuid.uuid4().hex}"
        schema = "uid UInt8, q Float64, v String, d DateTime"
        ds_response = await self.tb_api_proxy_async.create_datasource(
            new_shared_workspace_token,
            ds_name,
            schema
        )
        new_sharing_workspace = User.get_by_id(new_sharing_workspace.id)
        self.datasource = new_sharing_workspace.get_datasource(ds_response['datasource']['id'])
        self.wait_for_datasource_replication(new_sharing_workspace, self.datasource)

        # share Data Sources between them
        await self.tb_api_proxy_async.share_datasource_with_another_workspace(
            token=self.user_token,
            datasource_id=self.datasource.id,
            origin_workspace_id=new_sharing_workspace.id,
            destination_workspace_id=self.workspace.id,
            expect_notification=False
        )

        self.workspace = Users.get_by_id(self.workspace.id)
        self.assertEqual(len(self.workspace.get_datasources()), 1)
        self.datasource = self.workspace.get_datasources()[0]
        self.assertEqual(self.datasource.name, f"{new_sharing_workspace.name}.{ds_name}")

        # create pipe in workspace origin
        pipe_name = f"pipe_to_clone_{uuid.uuid4().hex}"
        await self.tb_api_proxy_async.create_pipe_endpoint(self.workspace, self.workspace_token, pipe_name,
                                                           "select count() from numbers(1,10)")
        self.workspace = User.get_by_id(self.workspace.id)
        self.workspace.get_pipe(pipe_name)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        await WorkspaceService.clone_datasources(self.workspace, new_workspace)
        await WorkspaceService.clone_pipes(self.workspace, new_workspace)

        self.workspace = User.get_by_id(self.workspace.id)
        new_workspace = User.get_by_id(new_workspace.id)

        workspace_diff = await WorkspaceService.compare_workspaces(self.workspace, new_workspace)
        self.assertEqual(workspace_diff.datasources, {'new': [], 'modified': [], 'deleted': [], 'only_metadata': []})
        self.assertEqual(workspace_diff.pipes, {'new': [], 'modified': [], 'deleted': [], 'only_metadata': []})

    @tornado.testing.gen_test
    async def test_clone_members(self):
        # new user account to add to workspace
        new_user = UserAccount.register(f'new_user_{uuid.uuid4().hex}@example.com', 'pass')
        self.users_to_delete.append(new_user)

        params = {
            'token': self.user_token,
            'operation': 'add',
            'users': new_user.email
        }
        url = f"/v0/workspaces/{self.workspace.id}/users?{urlencode(params)}"

        response = await self.fetch_async(url, method='PUT', body='')
        self.assertEqual(response.code, 200)
        result = json.loads(response.body)
        self.assertEqual(len(result['members']), 2)

        # create new workspace to clone
        workspace_name = f"ws_cloned_{uuid.uuid4().hex}"
        new_workspace = await self.tb_api_proxy_async.register_workspace(workspace_name, self.user_account,
                                                                         self.cluster)
        Users.get_token_for_scope(new_workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(new_workspace)

        await WorkspaceService.clone_members(self.workspace, new_workspace)
        new_workspace = User.get_by_id(new_workspace.id)
        self.workspace = User.get_by_id(self.workspace.id)

        workspace_members = UserWorkspaceRelationship.get_by_workspace(self.workspace.id)
        new_workspace_members = UserWorkspaceRelationship.get_by_workspace(new_workspace.id)

        self.assertEqual(len(workspace_members), len(new_workspace_members))
        for member in workspace_members:
            found = False
            for new_workspace_member in new_workspace_members:
                if member.user_id == new_workspace_member.user_id:
                    found = True
                    self.assertEqual(member.relationship, new_workspace_member.relationship)

            self.assertTrue(found, "Not all members found")



from datetime import timedelta, datetime
import uuid

from tests.views.base_test import BaseTest, TBApiProxy, STRIPE_SUBSCRIPTION_MOCK
from unittest.mock import patch
from tinybird.job import UsageRecordsTracker
from tinybird.token_scope import scopes
from tinybird.user import User, UserAccount


STRIPE_SUBSCRIPTIONS_MOCK = {'data': [STRIPE_SUBSCRIPTION_MOCK], "has_more": False}


class TestUsageRecordsTracker(BaseTest):
    def setUp(self):
        super(TestUsageRecordsTracker, self).setUp()
        self.tb_api_proxy = TBApiProxy(self)
        self.token = UserAccount.get_by_id(self.USER_ID).get_token_for_scope(scopes.AUTH)
        metrics_cluster = None

        self.usage_records_tracker = UsageRecordsTracker(metrics_cluster=metrics_cluster)

    @patch('stripe.Subscription.list')
    @patch('tinybird.plans.PlansService.track_usage_records', return_value=None)
    def test_track_records(self, plan_track_usage_records_mock, plan_subscriptions_list_mock):
        self.tb_api_proxy.create_customer(self.token, self.WORKSPACE_ID)
        self.tb_api_proxy.setup_payment_intent(self.token, self.WORKSPACE_ID)
        self.tb_api_proxy.subscribe_to_pro(self.token, self.WORKSPACE_ID)

        subscriptions = STRIPE_SUBSCRIPTIONS_MOCK
        subscriptions['data'][0]["metadata"]["workspace_id"] = self.WORKSPACE_ID
        plan_subscriptions_list_mock.return_value.auto_paging_iter.return_value = subscriptions['data']

        self.usage_records_tracker.track_usage_records()

        plan_track_usage_records_mock.assert_called_once_with(User.get_by_id(self.WORKSPACE_ID), STRIPE_SUBSCRIPTION_MOCK,
                                                              metrics_cluster=self.usage_records_tracker.metrics_cluster)

    @patch('stripe.Subscription.list')
    @patch('tinybird.plans.PlansService.track_usage_records', return_value=None)
    def test_do_not_track_records_if_subscription_from_other_region(self, plan_track_usage_records_mock,
                                                                    plan_subscriptions_list_mock):
        self.tb_api_proxy.create_customer(self.token, self.WORKSPACE_ID)
        self.tb_api_proxy.setup_payment_intent(self.token, self.WORKSPACE_ID)
        self.tb_api_proxy.subscribe_to_pro(self.token, self.WORKSPACE_ID)

        subscriptions = STRIPE_SUBSCRIPTIONS_MOCK
        subscriptions['data'][0]["created"] = datetime.timestamp(datetime.now() - timedelta(days=1))
        subscriptions['data'][0]["metadata"]["workspace_id"] = str(uuid.uuid4())
        plan_subscriptions_list_mock.return_value.auto_paging_iter.return_value = subscriptions['data']

        self.usage_records_tracker.track_usage_records()

        plan_track_usage_records_mock.assert_not_called()



import click
import os
import tornado
import requests
import tempfile
import shutil
from pathlib import Path

from urllib.parse import urlencode
from .views.base_test import BaseTest

from tinybird.ch import ch_get_replicas_for_table_sync
from tinybird.user import Users, User
from tinybird.token_scope import scopes

from tinybird.syncasync import sync_to_async
from tinybird.datafile import folder_push, folder_pull
from tinybird.client import TinyB

from io import StringIO


class TestProjects(BaseTest):

    def setUp(self):
        super().setUp()
        u = Users.get_by_id(self.WORKSPACE_ID)
        self.u = u
        self.admin_token = Users.get_token_for_scope(u, scopes.ADMIN)
        self.finished = False
        self.error = None

    async def __push_folder(self, folder, tag='', upload_fixtures=False, push_deps=True, filenames=None):
        self.finished = False
        self.error = None

        prev_cwd = os.getcwd()
        os.chdir(folder)

        tb_client = TinyB(self.admin_token, self.get_host())

        try:
            await folder_push(
                tb_client,
                folder=folder,
                force=True,
                push_deps=push_deps,
                tag=tag,
                upload_fixtures=upload_fixtures,
                filenames=filenames,
                wait=True)
        except Exception as e:
            self.error = e
        finally:
            os.chdir(prev_cwd)
            self.finished = True
        if self.error:
            raise self.error

    async def __pull_folder(self, folder, tag=None, match=None):
        self.finished = False
        self.error = None

        host = self.get_host()
        cl = TinyB(self.admin_token, host)
        try:
            await folder_pull(
                cl,
                folder=folder,
                auto=False,
                match=match,
                tag=tag,
                force=False)
        except Exception as e:
            self.error = e
        finally:
            self.finished = True

    async def push_data(self, name, data, mode='replace'):
        params = {
            'name': name,
            'token': self.admin_token,
            'mode': mode
        }
        create_url = f"/v0/datasources?{urlencode(params)}"
        post = sync_to_async(requests.post, thread_sensitive=False)
        r = await post(self.get_url(create_url), data=data)
        self.assertEqual(r.status_code, 200, r.content)
        return r

    async def _fetch(self, path, method='get', **kwargs):
        get = sync_to_async(getattr(requests, method.lower()), thread_sensitive=False)
        response = await get(self.get_url(path), **kwargs)
        return response

    @tornado.testing.gen_test
    async def test_join_tables(self):
        ch_get_replicas_for_table = sync_to_async(ch_get_replicas_for_table_sync, thread_sensitive=False)

        project_dir = os.path.dirname(__file__) + '/projects/test_view_to_join'
        await self.__push_folder(project_dir)

        # replace data
        s = StringIO("""a,b\n1,e\n2,d\n3,d\n4,d\n1,f""")
        await self.push_data('test', s)

        ds = Users.get_datasource(self.u, 'test')
        replicas = await ch_get_replicas_for_table(self.u.database_server, self.u.database, ds.id, self.u.cluster)
        is_ci = os.environ.get('CI', False)
        expected_min_number_of_replicas = 2 if is_ci else 1
        self.assertTrue(len(replicas) >= expected_min_number_of_replicas)
        for database_server in replicas:
            with User.transaction(self.u.id) as user:
                user.database_server = database_server
            params = {
                'q': "select joinGet('test_join', 's', 'd') as key format JSON",
                'token': self.admin_token
            }
            response = await self._fetch(f'/v0/sql?{urlencode(params)}')
            self.assertEqual(response.status_code, 200, response.content)
            result = response.json()
            self.assertEqual(result['data'][0], {'key': 9}, f'{database_server} is not replicated')

    @tornado.testing.gen_test
    async def test_fail_to_parse_tsv(self):
        await self.__push_folder(os.path.dirname(__file__) + '/projects/test_fail_to_parse_tsv', tag='dev', upload_fixtures=True)
        params = {
            'q': "select count() c from dev__wikipedia_quarantine",
            'token': self.admin_token
        }
        response = await self._fetch(f'/v0/sql?{urlencode(params)}')
        result = response.json()
        self.assertEqual(result, 4, result)

    @tornado.testing.gen_test
    async def test_should_upload_last_versions_from_scracth(self):
        await self.__push_folder(os.path.dirname(__file__) + '/projects/test_versions')

        params = {
            'token': self.admin_token
        }
        response = await self._fetch(f'/v0/datasources/test__v1?{urlencode(params)}')
        self.assertEqual(response.status_code, 200, response)
        response = await self._fetch(f'/v0/pipes/transform_data__v0?{urlencode(params)}')
        self.assertEqual(response.status_code, 200, response)
        response = await self._fetch(f'/v0/pipes/test_endpoint__v1?{urlencode(params)}')
        self.assertEqual(response.status_code, 200, response)

    @tornado.testing.gen_test
    async def test_should_use_project_versions(self):

        # create a new version and push the project
        params = {
            'token': self.admin_token,
            'name': 'test__v2',
            'schema': 'a Int32, b Float64, c String'
        }
        response = await self._fetch(f'/v0/datasources/?{urlencode(params)}', method='POST', data='')

        await self.__push_folder(os.path.dirname(__file__) + '/projects/test_versions')
        params = {
            'token': self.admin_token
        }
        response = await self._fetch(f'/v0/pipes/test_endpoint__v1?{urlencode(params)}')
        r = response.json()
        # here it must use v1 because that's what we have in the project
        self.assertEqual('test__v1' in r['nodes'][0]['sql'], True, r)

        response = await self._fetch(f'/v0/pipes/transform_data__v0?{urlencode(params)}')
        r = response.json()
        # here it must use v1 because that's what we have in the project
        self.assertEqual('test__v1' in r['nodes'][0]['sql'], True, r)

    @tornado.testing.gen_test
    async def test_should_push_only_one_file(self):
        params = {
            'token': self.admin_token,
            'name': 'ony_one_file__test__v1',
            'schema': 'a Int32, b Float64, c String'
        }
        response = await self._fetch(f'/v0/datasources/?{urlencode(params)}', method='POST', data='')

        folder_path = os.path.dirname(__file__) + '/projects/test_versions'
        await self.__push_folder(
            folder_path,
            filenames=[folder_path + "/endpoints/test_endpoint.pipe"],
            tag='ony_one_file',
            push_deps=False
        )
        params = {
            'token': self.admin_token
        }
        response = await self._fetch(f'/v0/pipes/ony_one_file__test_endpoint__v1?{urlencode(params)}')
        self.assertEqual(response.status_code, 200, response)
        r = response.json()
        # here it must use v1 because that's what we have in the project (even if it's not pushed)
        self.assertEqual('ony_one_file__test__v1' in r['nodes'][0]['sql'], True, r)

    @tornado.testing.gen_test
    async def test_should_upload_last_versions_with_tag(self):
        await self.__push_folder(os.path.dirname(__file__) + '/projects/test_versions', tag='mybranch')

        params = {
            'token': self.admin_token
        }
        response = await self._fetch(f'/v0/datasources/mybranch__test__v1?{urlencode(params)}')
        self.assertEqual(response.status_code, 200, response)
        response = await self._fetch(f'/v0/pipes/mybranch__transform_data__v0?{urlencode(params)}')
        self.assertEqual(response.status_code, 200, response)
        response = await self._fetch(f'/v0/pipes/mybranch__test_endpoint__v1?{urlencode(params)}')
        self.assertEqual(response.status_code, 200, response)
        response = await self._fetch(f'/v0/datasources/test__v1?{urlencode(params)}')
        self.assertEqual(response.status_code, 404, response)

    @tornado.testing.gen_test
    async def test_should_pull_last_versions(self):
        params = {
            'token': self.admin_token,
            'name': 'test__v2',
            'schema': 'a Int32, b Float64, c String'
        }
        await self._fetch(f'/v0/datasources/?{urlencode(params)}', method='POST', data='')

        params = {
            'token': self.admin_token,
            'name': 'test_pipe__v1',
            'sql': 'select count() from test__v2'
        }
        await self._fetch(f'/v0/pipes/?{urlencode(params)}', method='POST', data='')

        tmppath = tempfile.mkdtemp()
        await self.__pull_folder(tmppath)
        project_folder = Path(tmppath)
        ds = project_folder / 'test.datasource'
        self.assertEqual(ds.exists(), True, list(project_folder.iterdir()))
        with open(ds) as file:
            contents = file.read()
        self.assertEqual('VERSION 2' in contents, True, contents)

        ds = project_folder / 'test_pipe.pipe'
        with open(ds) as file:
            contents = file.read()
        self.assertEqual(ds.exists(), True, list(project_folder.iterdir()))
        self.assertEqual('VERSION 1' in contents, True, contents)
        shutil.rmtree(tmppath)

    @tornado.testing.gen_test
    async def test_should_use_server_missing_deps(self):
        # create a datasource which is missing locally so in the push is used
        params = {
            'token': self.admin_token,
            'name': 'test_missing_locally__v1',
            'schema': 'a Int32, b Float64, c String'
        }
        await self._fetch(f'/v0/datasources/?{urlencode(params)}', method='POST', data='')

        await self.__push_folder(os.path.dirname(__file__) + '/projects/test_missing_deps')

        params = {
            'token': self.admin_token
        }
        response = await self._fetch(f'/v0/pipes/test_endpoint__v1?{urlencode(params)}')
        self.assertEqual(response.status_code, 200, response)
        r = response.json()
        # here it must use v1 because that's what we have in the project (even if it's not pushed)
        self.assertEqual('test_missing_locally__v1' in r['nodes'][0]['sql'], True, r)

    @tornado.testing.gen_test
    async def test_should_push_matview_and_check_engine(self):
        await self.__push_folder(os.path.dirname(__file__) + '/projects/test_mat_views')

    @tornado.testing.gen_test
    async def test_datasource_deps(self):
        with self.assertRaisesRegex(Exception, r"Resource \'worldwide_stores_join_by_country\' not found"):
            await self.__push_folder(os.path.dirname(__file__) + '/projects/ds_deps', filenames=['sales.datasource'], push_deps=False)

        await self.__push_folder(os.path.dirname(__file__) + '/projects/ds_deps', filenames=['sales.datasource'], push_deps=True)
        params = {
            'token': self.admin_token
        }
        for ds in ('sales', 'stores_join', 'worldwide_stores_join_by_country'):
            response = await self._fetch(f'/v0/datasources/{ds}?{urlencode(params)}')
            self.assertEqual(response.status_code, 200, response)

    @tornado.testing.gen_test
    async def test_wrong_engine_vars(self):
        with self.assertRaisesRegex(Exception, r"ENGINE_WRONG_KEY is not a valid option"):
            await self.__push_folder(os.path.dirname(__file__) + '/projects/test_datafiles', filenames=['wrong_engine_var.datasource'], push_deps=False)

    @tornado.testing.gen_test
    async def test_sql_injection_in_engine(self):
        with self.assertRaisesRegex(Exception, r"Failed creating Data Source: Invalid data source structure: CREATE TABLE: Unsupported CREATE AS query"):
            await self.__push_folder(os.path.dirname(__file__) + '/projects/test_datafiles',
                                     filenames=['engine_sql_injection.datasource'], push_deps=False)

    @tornado.testing.gen_test
    async def test_sql_injection_in_materialized_column(self):
        with self.assertRaisesRegex(click.exceptions.ClickException,
                                    r"DB::Exception: Usage of function remote is restricted"):
            await self.__push_folder(os.path.dirname(__file__) + '/projects/test_datafiles',
                                     filenames=['materialized_sql_injection.datasource'], push_deps=False)



import asyncio
from chtoolset import query as chquery
from collections import Counter
import json
import threading
import time
import uuid
import zlib
import unittest

from functools import partial
from http.server import SimpleHTTPRequestHandler, HTTPServer
from io import BytesIO
from socketserver import ThreadingMixIn
from unittest.mock import Mock, call, patch, ANY

import requests
from tests.conftest import CH_ADDRESS, get_clickhouse_version, OTHER_STORAGE_POLICY
from tinybird.constants import CHCluster

from tinybird.syncasync import async_to_sync
from tinybird.sql import parse_table_structure
from tinybird.user import User, UserAccount
from tinybird.app import global_patch_for_requests_timeout
from tinybird.ch import (
    ch_create_exact_empty_copy, ch_guarded_query as _ch_guarded_query,
    HTTPClient, ch_table_details_async, ch_truncate_table,
    ch_truncate_table_with_fallback,
    url_from_host, ch_table_partitions_sync,
    ch_wait_for_query_cancellation, WaitingCancelOperationResult,
    table_structure, CHTable, ch_describe_query,
    ch_get_columns_from_query, CSVInfo, ch_table_dependent_views_sync,
    CHTableLocation, ch_table_partitions_for_sample_sync,
    _get_query_log, ch_flush_logs_on_all_replicas, ch_get_cluster_instances,
    ch_flush_logs_on_all_replicas_sync, ch_attach_partitions, ch_query_table_partitions, Partitions)
from tinybird.csv_tools import csv_from_python_object
from tinybird.ch_utils.exceptions import CHException
from tinybird.ch_utils.errors import CHErrors

ch_guarded_query = partial(_ch_guarded_query, check_frequency_seconds=.05)
sync_ch_describe_query = async_to_sync(ch_describe_query)


class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
    daemon_threads = True
    first_query_log_query = True


class ProxyHandler(SimpleHTTPRequestHandler):
    def __init__(self, database_server, *args, **kwargs):
        self.upstream_server = url_from_host(database_server)[:-1]
        super().__init__(*args, **kwargs)

    def log_message(self, *args, **kwargs):
        pass

    def do_GET(self):
        if self.server.first_query_log_query and 'fail_first_query_log_query' in self.path:
            self.server.first_query_log_query = False
            self.send_response(500)
            self.end_headers()
            self.copyfile(BytesIO('Code: 160, e.displayText() = DB::Exception: Estimated query execution time (22.506430934551624 seconds) is too long. Maximum: 10. Estimated rows to process: 159418181: While executing MergeTreeThread (version 20.7.2.30 (official build))'.encode()), self.wfile)
            return
        r = requests.get(f"{self.upstream_server}{self.path}")
        self.send_response(r.status_code)
        self.end_headers()
        self.copyfile(BytesIO(r.content), self.wfile)

    def do_POST(self):
        query_length = int(self.headers.get('Content-Length', 0))
        query = self.rfile.read(query_length)
        try:
            # timeout=(0.5, 0.5)
            # We could define a timeout to better emulate the Varnish behaviour.
            # However, requests close the connection to the actual ClickHouse server
            # and that makes ClickHouse unhappy, logging an error about not being able
            # to write to the output stream.
            _ = requests.post(f"{self.upstream_server}{self.path}", data=query)
            # self.send_response(r.status_code)
            # self.end_headers()
            # self.copyfile(BytesIO(r.content), self.wfile)
        finally:
            self.send_response(503)
            self.end_headers()


class TestGuardedQueries(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        WORKSPACE = f'test_ch_{uuid.uuid4().hex}'
        USER = f'{WORKSPACE}@example.com'

        cls.user = UserAccount.register(USER, 'pass')
        cls.workspace = User.register(WORKSPACE, cls.user.id, CHCluster(name='tinybird', server_url=CH_ADDRESS))
        client = HTTPClient(cls.workspace.database_server, database=None)
        client.query_sync(f"CREATE DATABASE IF NOT EXISTS {cls.workspace.database} ON CLUSTER tinybird", read_only=False)

    @classmethod
    def tearDownClass(cls):
        database = cls.workspace.database
        client = HTTPClient(cls.workspace.database_server, database=None)
        User._delete(cls.workspace.id)
        UserAccount._delete(cls.user.id)
        client.query_sync(f"DROP DATABASE IF EXISTS `{database}` ON CLUSTER tinybird SYNC", read_only=False)
        super().tearDownClass()

    def test_query_ok(self):
        query = "SELECT 1"
        query_id, query_finish = ch_guarded_query(self.workspace.database_server, self.workspace.database, query, 'tinybird')
        self.assertIsNotNone(query_id)
        self.assertIsNotNone(query_finish)

    def test_query_error(self):
        query = "SELECT faulty query"
        with self.assertRaisesRegex(CHException, r"Missing columns: 'faulty'"):
            ch_guarded_query(self.workspace.database_server, self.workspace.database, query, 'tinybird')

    def test_query_max_execution_time(self):
        query = "SELECT sleepEachRow(1) FROM numbers(3)"
        with self.assertRaisesRegex(CHException, r"Timeout exceeded"):
            ch_guarded_query(self.workspace.database_server, self.workspace.database, query, 'tinybird', max_execution_time=1)

    def test_varnish_behaviour(self):
        query = "SELECT sleep(0.01)"
        query_id, query_finish = ch_guarded_query(CH_ADDRESS, self.workspace.database, query, 'tinybird')
        self.assertIsNotNone(query_id)
        self.assertIsNotNone(query_finish)

    def test_query_ok_with_failure_on_check(self):
        query = "SELECT 1"
        query_id, query_finish = ch_guarded_query(CH_ADDRESS, self.workspace.database, query, 'tinybird',
                                                  query_id="fail_first_query_log_query")
        self.assertIsNotNone(query_id)
        self.assertIsNotNone(query_finish)

    def test_query_directly_returns_if_externally_cancelled(self):

        externally_cancelled_function = Mock(return_value=True)

        query = "SELECT sleep(3)"
        query_id, query_finish = ch_guarded_query(
            self.workspace.database_server, self.workspace.database, query, 'tinybird',
            query_id="random_query_id", has_been_externally_cancelled=externally_cancelled_function)
        self.assertIsNotNone(query_id)
        self.assertIsNone(query_finish)

    def test_query_returns_after_the_query_is_made_if_externally_cancelled(self):

        externally_cancelled_function = Mock(side_effect=[False, True, True])

        query = "SELECT sleep(0.01)"
        query_id, query_finish = ch_guarded_query(
            self.workspace.database_server, self.workspace.database, query, 'tinybird', query_id="random_query_id",
            has_been_externally_cancelled=externally_cancelled_function)
        self.assertIsNotNone(query_id)
        self.assertIsNone(query_finish)
        externally_cancelled_function.assert_has_calls([call()] * 2)


class TestPartitions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        WORKSPACE = f'test_ch_{uuid.uuid4().hex}'
        USER = f'{WORKSPACE}@example.com'
        cls.user = UserAccount.register(USER, 'pass')
        cls.workspace = User.register(WORKSPACE, cls.user.id, CHCluster(name='tinybird', server_url=CH_ADDRESS))
        client = HTTPClient(cls.workspace.database_server, database=None)
        client.query_sync(f"CREATE DATABASE IF NOT EXISTS {cls.workspace.database} ON CLUSTER tinybird", read_only=False)

    @classmethod
    def tearDownClass(cls):
        client = HTTPClient(cls.workspace.database_server, database=None)
        database = cls.workspace.database
        User._delete(cls.workspace.id)
        UserAccount._delete(cls.user.id)
        client.query_sync(f"DROP DATABASE IF EXISTS {database} ON CLUSTER tinybird SYNC", read_only=False)
        super().tearDownClass()

    def _create_table(self, name, schema, partition_key, data_sql, engine: str = 'MergeTree'):
        client = HTTPClient(self.workspace.database_server, database=self.workspace.database)
        client.query_sync(f"DROP TABLE IF EXISTS {name} ON CLUSTER tinybird SYNC", read_only=False)
        full_engine = "Replicated" + engine + "('/clickhouse/tables/{layer}-{shard}/" + name + "','{replica}')"
        create_sql = f'''CREATE TABLE {name} ON CLUSTER tinybird ({schema})
            ENGINE = {full_engine}
            PARTITION BY {partition_key}
            ORDER BY tuple()'''
        client.query_sync(create_sql, read_only=False)
        if data_sql:
            client.query_sync(f"INSERT INTO {name} {data_sql}", read_only=False)

    def test_ch_partitions_tuple(self):
        schema = 'a String, b DateTime'
        partition_key = 'tuple()'
        data_sql = """SELECT
            'a' as a,
            now() as b
        FROM numbers(10)"""
        self._create_table("test_ch_partitions_tuple", schema, partition_key, data_sql)

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database, ["test_ch_partitions_tuple"])
        self.assertEqual(partitions, ['tuple()'])

    def test_ch_partitions_date(self):
        schema = 'a String, b DateTime'
        partition_key = 'toDate(b)'
        data_sql = """SELECT
            'a' as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(5)"""
        self._create_table("test_ch_partitions_date", schema, partition_key, data_sql)

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database, ["test_ch_partitions_date"])
        self.assertEqual(
            partitions,
            ["'2020-12-05'", "'2020-12-04'", "'2020-12-03'", "'2020-12-02'", "'2020-12-01'"]
        )

    def test_ch_partitions_subset(self):
        schema = 'a String, b DateTime'
        partition_key = 'substring(a, 1, 1)'
        data_sql = """SELECT
            toString(number) as a,
            now() as b
        FROM numbers(1000)"""
        self._create_table("test_ch_partitions_subset", schema, partition_key, data_sql)

        partitions = ch_table_partitions_for_sample_sync(self.workspace.database_server, self.workspace.database, "test_ch_partitions_subset", max_rows=10)
        self.assertEqual(len(partitions), 1)
        # least(1000 * 0.1, 10) => 10
        self.assertEqual(partitions[0][1], 10)

    def test_ch_partitions_yyyymm(self):
        schema = 'a String, b DateTime'
        partition_key = 'toYYYYMM(b)'
        data_sql = """SELECT
            'a' as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(90)"""
        self._create_table("test_ch_partitions_yyyymm", schema, partition_key, data_sql)

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database, ["test_ch_partitions_yyyymm"])
        self.assertEqual(partitions, ['202102', '202101', '202012'])

    def test_ch_partitions_numeric(self):
        schema = 'n Int8, b DateTime'
        partition_key = 'n % 5'
        data_sql = """SELECT
            number as n,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(25)"""
        self._create_table("test_ch_partitions_numeric", schema, partition_key, data_sql)

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database, ["test_ch_partitions_numeric"])
        self.assertEqual(partitions, ['4', '3', '2', '1', '0'])

    def test_ch_partitions_two_partition_keys(self):
        schema = 'a String, b DateTime'
        partition_key = '(a, toYYYYMM(b))'
        data_sql = """SELECT
            if(number % 2 = 0, 'a', 'b') as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(90)"""
        self._create_table("test_ch_partitions_two_partition_keys", schema, partition_key, data_sql)

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database, ["test_ch_partitions_two_partition_keys"])
        self.assertEqual(
            Counter(partitions),
            Counter(["('b',202102)", "('b',202101)", "('b',202012)", "('a',202102)", "('a',202101)", "('a',202012)"])
        )

    def test_ch_partitions_datetime(self):
        schema = 'a String, b DateTime'
        partition_key = 'b'
        data_sql = """SELECT
            'a' as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(3)"""
        self._create_table("test_ch_partitions_datetime", schema, partition_key, data_sql)

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database, ["test_ch_partitions_datetime"])
        self.assertEqual(
            partitions,
            ["'2020-12-03 00:00:00'", "'2020-12-02 00:00:00'", "'2020-12-01 00:00:00'"]
        )

    def test_ch_partitions_with_met_condition(self):
        schema = 'a String, b DateTime'
        partition_key = 'toYYYYMM(b)'
        data_sql = """SELECT
            'a' as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(90)"""
        self._create_table("test_ch_partitions_with_met_condition", schema, partition_key, data_sql)
        condition = "b > toDateTime('2021-01-01 00:00:00') and a = 'a'"
        expected_partitions = ['202102', '202101']

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database,
                                              ["test_ch_partitions_with_met_condition"], condition)
        self.assertEqual(partitions, expected_partitions)

    def test_ch_partitions_with_unmet_condition(self):
        schema = 'a String, b DateTime'
        partition_key = 'toYYYYMM(b)'
        data_sql = """SELECT
            'a' as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(90)"""
        self._create_table("test_ch_partitions_with_unmet_condition", schema, partition_key, data_sql)
        condition = "b > toDateTime('2023-01-01 00:00:00') and a = 'a'"
        expected_partitions = []

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database,
                                              ["test_ch_partitions_with_unmet_condition"], condition)
        self.assertEqual(partitions, expected_partitions)

    def test_ch_partitions_with_unknown_column_in_condition(self):
        schema = 'a String, b DateTime'
        partition_key = 'toYYYYMM(b)'
        data_sql = """SELECT
            'a' as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(90)"""
        self._create_table("test_ch_partitions_with_unknown_column_in_condition", schema, partition_key, data_sql)
        condition = "whatever > toDateTime('2023-01-01 00:00:00')"
        expected_partitions = ['202102', '202101', '202012']

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database,
                                              ["test_ch_partitions_with_unknown_column_in_condition"], condition)
        self.assertEqual(partitions, expected_partitions)

    def test_ch_partitions_with_other_table_in_condition(self):
        schema = 'a String, b DateTime'
        partition_key = 'toYYYYMM(b)'
        data_sql = """SELECT
            'a' as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(90)"""
        self._create_table("test_ch_partitions_with_other_table_in_condition", schema, partition_key, data_sql)

        other_table = 'other_table'
        other_engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + other_table + "','{replica}')"
        create_sql = f'''CREATE TABLE {other_table} ON CLUSTER tinybird (a String, b DateTime) ENGINE = {other_engine} ORDER BY b'''
        client = HTTPClient(self.workspace.database_server, database=self.workspace.database)
        client.query_sync(create_sql, read_only=False)
        data_sql = """SELECT
            'a' as a,
            toDateTime('2021-01-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(90)"""
        client.query_sync(f"INSERT INTO {other_table} {data_sql}", read_only=False)
        condition = f"b > (SELECT min(b) FROM {self.workspace.database}.{other_table})"

        expected_partitions = ['202102', '202101']

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database,
                                              ["test_ch_partitions_with_other_table_in_condition"], condition)
        self.assertEqual(partitions, expected_partitions)

    @patch('tinybird.ch._query_table_partitions_with_condition_fallback_sync')
    def test_ch_partitions_for_views(self, mocked_partition_query):
        # The partition query should not be called for Views.
        # Views are not supported on the product, and this is only needed because of the usage of Views on ITX.
        test_name = f'test_ch_partitions_for_views_{uuid.uuid4().hex}'
        table_name, view_name = f'{test_name}_table', f'{test_name}_view'

        schema = 'a String, b DateTime'
        partition_key = 'toYYYYMM(b)'
        data_sql = """SELECT
            'a' as a,
            toDateTime('2020-12-01 00:00:00') + INTERVAL number DAY as b
        FROM numbers(90)"""
        self._create_table(table_name, schema, partition_key, data_sql, engine='ReplacingMergeTree')

        client = HTTPClient(self.workspace.database_server, database=self.workspace.database)
        create_sql = f'''
        CREATE VIEW IF NOT EXISTS {test_name} ON CLUSTER tinybird ({schema})
        AS SELECT * FROM {table_name}'''
        client.query_sync(create_sql, read_only=False)

        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database, [view_name])
        self.assertEqual(partitions, [])
        condition = "b > toDateTime('2021-01-01 00:00:00') and a = 'a'"
        partitions = ch_table_partitions_sync(self.workspace.database_server, self.workspace.database, [view_name], condition)
        self.assertEqual(partitions, [])

        mocked_partition_query.assert_not_called()

    def test_ch_attach_string_partitions(self):
        schema = 'a String'
        partition_key = 'substring(a, 1, 1)'
        self._create_table("test_ch_partitions_substring", schema, partition_key, None)
        self._create_table("test_ch_partitions_substring2", schema, partition_key, None)
        client = HTTPClient(self.workspace.database_server, database=self.workspace.database)
        client.query_sync("INSERT INTO test_ch_partitions_substring values(' test')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_substring values('(')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_substring values(')')", read_only=False)
        # uncomment after we are in CH 22
        # client.query_sync("INSERT INTO test_ch_partitions_substring values('123')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_substring values('.-')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_substring values('/')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_substring values('''')", read_only=False)

        partitions = async_to_sync(ch_query_table_partitions)(self.workspace.database_server, [self.workspace.database])
        partitions = [p for p in partitions if p.table == 'test_ch_partitions_substring']
        partitions = Partitions().as_sql_list(partitions)
        _, result = async_to_sync(ch_attach_partitions)(self.workspace.database_server, self.workspace.database, "test_ch_partitions_substring2", "test_ch_partitions_substring", partitions)
        self.assertEqual(result, b'')

        schema = 'a String'
        partition_key = 'a'
        self._create_table("test_ch_partitions_substring3", schema, partition_key, None)
        self._create_table("test_ch_partitions_substring4", schema, partition_key, None)
        client = HTTPClient(self.workspace.database_server, database=self.workspace.database)
        client.query_sync("INSERT INTO test_ch_partitions_substring values(' test')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_substring values('.-')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_substring values('/')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_substring values('''')", read_only=False)

        partitions = async_to_sync(ch_query_table_partitions)(self.workspace.database_server, [self.workspace.database])
        partitions = [p for p in partitions if p.table == 'test_ch_partitions_substring']
        partitions = Partitions().as_sql_list(partitions)
        _, result = async_to_sync(ch_attach_partitions)(self.workspace.database_server, self.workspace.database, "test_ch_partitions_substring2", "test_ch_partitions_substring", partitions)
        self.assertEqual(result, b'')

    def test_ch_attach_compound_partitions(self):
        schema = 'a String, b DateTime'
        partition_key = '(toYYYYMM(b), a)'
        self._create_table("test_ch_partitions_compound", schema, partition_key, None)
        self._create_table("test_ch_partitions_compound2", schema, partition_key, None)
        client = HTTPClient(self.workspace.database_server, database=self.workspace.database)
        client.query_sync("INSERT INTO test_ch_partitions_compound values(' test', '2020-01-01 00:00:00')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_compound values('.-', '2021-01-01 00:00:00')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_compound values('/', '2022-01-01 00:00:00')", read_only=False)
        client.query_sync("INSERT INTO test_ch_partitions_compound values('''', '2023-01-01 00:00:00')", read_only=False)

        partitions = async_to_sync(ch_query_table_partitions)(self.workspace.database_server, [self.workspace.database])
        partitions = [p for p in partitions if p.table == 'test_ch_partitions_compound']
        partitions = Partitions().as_sql_list(partitions)
        _, result = async_to_sync(ch_attach_partitions)(self.workspace.database_server, self.workspace.database, "test_ch_partitions_compound2", "test_ch_partitions_compound", partitions)
        self.assertEqual(result, b'')


class TestCancellableQueries(unittest.IsolatedAsyncioTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.database_host = User.default_database_server
        cls.test_database = f'd_cancellable_{uuid.uuid4().hex}'

        ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        ch_client_for_setup.query_sync(f"CREATE DATABASE IF NOT EXISTS {cls.test_database} ON CLUSTER tinybird", read_only=False)

        cls.client = HTTPClient(cls.database_host, database=cls.test_database)

    @classmethod
    def tearDownClass(cls):
        ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {cls.test_database} ON CLUSTER tinybird SYNC", read_only=False)
        super().tearDownClass()

    async def test_trying_to_cancel_a_query_that_is_not_actually_running_returns_not_found(self):

        query_id = str(uuid.uuid4())

        result = await ch_wait_for_query_cancellation(
            self.database_host, self.test_database, query_id, 'tinybird', check_frequency_seconds=0.05)

        assert result == WaitingCancelOperationResult.NOT_FOUND

    async def test_trying_to_cancel_a_query_operation_can_finish(self):
        query_id = str(uuid.uuid4())

        def f():
            try:
                self.client.query_sync("SELECT sum(number) FROM numbers(4000000000) settings max_threads = 1;",
                                       query_id=query_id)
            except Exception:
                pass

        self.parallel_thread_to_run_the_query = threading.Thread(target=f)
        self.parallel_thread_to_run_the_query.start()

        await wait_for_query_start(self.client, query_id)

        result = await ch_wait_for_query_cancellation(
            self.database_host, self.test_database, query_id, 'tinybird', wait_seconds=60, check_frequency_seconds=0.05)

        assert result == WaitingCancelOperationResult.CANCELLED or result == WaitingCancelOperationResult.NOT_FOUND


class TestCancellableQueriesBatch2(unittest.IsolatedAsyncioTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.database_host = User.default_database_server
        cls.test_database = f'd_cancellable_{uuid.uuid4().hex}'

        ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        ch_client_for_setup.query_sync(f"CREATE DATABASE IF NOT EXISTS {cls.test_database} ON CLUSTER tinybird", read_only=False)

        cls.client = HTTPClient(cls.database_host, database=cls.test_database)

    @classmethod
    def tearDownClass(cls):
        ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {cls.test_database} ON CLUSTER tinybird SYNC", read_only=False)
        super().tearDownClass()

    async def test_trying_to_cancel_a_query_using_async_operation_raise_exception_if_timeout_is_exceeded(self):
        def f():
            try:
                self.client.query_sync("SELECT sleepEachRow(1) from numbers(3);",
                                       query_id=query_id)
            except Exception:
                pass

        query_id = str(uuid.uuid4())
        self.parallel_thread_to_run_the_query = threading.Thread(target=f)
        self.parallel_thread_to_run_the_query.start()

        await wait_for_query_start(self.client, query_id)

        with self.assertRaisesRegex(CHException, r"Timeout exceeded"):
            await ch_wait_for_query_cancellation(self.database_host, self.test_database, query_id, 'tinybird',
                                                 check_frequency_seconds=0.05, wait_seconds=0.01)


class TestStringTableStructureMapping(unittest.TestCase):

    def test_with_different_column_features(self):
        schema = "`simple_column` String, " \
                 "`colummn_with_nullable` Nullable(Int16), " \
                 "`colummn_with_nullable_default_null` Nullable(Int16) DEFAULT NULL, " \
                 "`column_with_default` String DEFAULT 'bla'," \
                 "`column_materialized` String materialized upper(simple_column)"

        columns = parse_table_structure(schema)

        structure = table_structure(columns)

        self.assertEqual("`simple_column` String, "
                         "`colummn_with_nullable` Nullable(Int16), "
                         "`colummn_with_nullable_default_null` Nullable(Int16), "
                         "`column_with_default` String DEFAULT 'bla', "
                         "`column_materialized` String MATERIALIZED upper(simple_column)",
                         structure)

    def test_columns_without_default_value_and_with_name_to_normalize(self):
        columns = [
            {'name': 'my_date_column', 'normalized_name': 'my_date_column', 'type': 'Date', 'nullable': False, 'auto': False, 'default_value': None, 'codec': None},
            {'name': 'other_column', 'normalized_name': 'other_column', 'type': 'Int16', 'nullable': False, 'auto': False, 'default_value': None, 'codec': None}
        ]

        structure = table_structure(columns)

        self.assertEqual("`my_date_column` Date, `other_column` Int16", structure)


class TestCreateTableGeneration(unittest.TestCase):
    def test_generate_create_table_with_nullable_columns(self):

        schema = """
               `no_nullable_string` String,
               `nullable_column` Nullable(Int64),
               `nullable_column_with_default_null` Nullable(Int64) DEFAULT NULL,
               `nullable_column_with_default_cast_null` Nullable(Int64) DEFAULT CAST(NULL, 'Nullable(Int64)'),
               `no_nullable_with_default` String DEFAULT 'bla',
               `no_nullable_with_default_and_cast` String DEFAULT CAST(123, 'String')
            """

        columns = parse_table_structure(schema)

        ch_table = CHTable(columns, cluster='tinybird')

        create_table_statement = ch_table.as_sql("thedatabase", "thetable")
        expected_table_statement = """CREATE TABLE thedatabase.thetable ON CLUSTER tinybird
(
    `no_nullable_string` String,
    `nullable_column` Nullable(Int64),
    `nullable_column_with_default_null` Nullable(Int64),
    `nullable_column_with_default_cast_null` Nullable(Int64) DEFAULT CAST(NULL, 'Nullable(Int64)'),
    `no_nullable_with_default` String DEFAULT 'bla',
    `no_nullable_with_default_and_cast` String DEFAULT CAST(123, 'String')
)
ENGINE = ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/thedatabase.thetable', '{replica}')
ORDER BY (no_nullable_string, no_nullable_with_default, no_nullable_with_default_and_cast)"""

        self.assertEqual(chquery.format(expected_table_statement), chquery.format(create_table_statement))  # noqa: W291


class TestSqlAnalysisAndValidations(unittest.IsolatedAsyncioTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.database_host = User.default_database_server
        cls.test_database = f'd_validation_{uuid.uuid4().hex}'

        cls.ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        cls.ch_client_for_setup.query_sync(f"CREATE DATABASE IF NOT EXISTS {cls.test_database} ON CLUSTER tinybird", read_only=False)

        cls.client = HTTPClient(cls.database_host, database=cls.test_database)

    @classmethod
    def tearDownClass(cls):
        cls.ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {cls.test_database} ON CLUSTER tinybird SYNC", read_only=False)
        super().tearDownClass()

    def test_describe_query_returns_ok_with_correct_sql(self):
        headers, returned_query = sync_ch_describe_query(self.database_host, self.test_database, "select 1")
        self.assertEqual(returned_query, b"1\tUInt8\t\t\t\t\t\n")
        self.assertEqual(headers['X-Http-Reason'], 'OK')

    def test_describe_different_expressions_with_same_alias_error(self):
        sql = '''
            SELECT
                property_id,
                countMerge(bookings_visits) AS total_bookings
            FROM table_2
            WHERE ((visit_date >= parseDateTimeBestEffort('2021-05-01')) AND (visit_date <= parseDateTimeBestEffort('2021-05-31 23:59:59'))) AND (property_id IN
            (
                    SELECT property_id
                FROM
                (
                        SELECT *
                    FROM table_1
                    WHERE (parseDateTimeBestEffort('2021-05-01') >= min_visit_date) AND (max_visit_date >= (parseDateTimeBestEffort('2021-05-31') - 2)) AND (compset_property_id IN (1000060))
                ) AS t_057a8d989fc0465a92e3668db0e75785
            ))
            GROUP BY property_id
        '''

        table_name = 'table_1'
        engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name + "','{replica}')"
        self.client.query_sync(f"""
            CREATE TABLE {table_name} ON CLUSTER tinybird (key String, property_id Int64, compset_property_id Int64, min_visit_date Date, max_visit_date Date) ENGINE = {engine} Order by compset_property_id
        """, read_only=False)

        engine_2 = "ReplicatedAggregatingMergeTree('/clickhouse/tables/{layer}-{shard}/table_2','{replica}')"
        self.client.query_sync(f"""
            CREATE TABLE table_2 ON CLUSTER tinybird (visit_date Date, property_id Int64, `bookings_visits` AggregateFunction(count, String)) ENGINE = {engine_2} Partition By toYYYYMM(visit_date) Order by (visit_date, property_id)
        """, read_only=False)

        headers, returned_query = sync_ch_describe_query(self.database_host, self.test_database, sql)
        self.assertEqual(returned_query, b"property_id\tInt64\t\t\t\t\t\ntotal_bookings\tUInt64\t\t\t\t\t\n")
        self.assertEqual(headers['X-Http-Reason'], 'OK')

    def test_describe_query_fails_with_incorrect_sql(self):
        with self.assertRaises(CHException) as context:
            sync_ch_describe_query(self.database_host, self.test_database, "select 1 from x where")

        self.assertEqual(CHErrors.SYNTAX_ERROR, context.exception.code)

    def test_describe_query_fails_with_correct_sql_but_column_not_found(self):
        table_name = "example_table"
        engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name + "','{replica}')"
        self.client.query_sync(f"""
            CREATE TABLE {table_name} ON CLUSTER tinybird (column_a String, column_b String) ENGINE = {engine} Order by tuple()
        """, read_only=False)

        with self.assertRaises(CHException) as context:
            sync_ch_describe_query(
                self.database_host, self.test_database, "SELECT column_a, column_b, column_c FROM example_table")

        self.assertEqual(CHErrors.UNKNOWN_IDENTIFIER, context.exception.code)
        self.assertRegex(context.exception.args[0],
                         r"\[Error\] Missing columns: 'column_c' while processing query: 'SELECT "
                         "column_a, column_b, column_c FROM example_table', required columns: "
                         "'column_a' 'column_b' 'column_c'.*",
                         )

    def test_too_many_parts_error_is_correctly_parsed(self):
        table_name = f"too_many_parts_table_{uuid.uuid4().hex}"
        engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name + "','{replica}')"
        self.client.query_sync(f"""
            CREATE TABLE {table_name} ON CLUSTER tinybird (d Date, n Int32) ENGINE = {engine} PARTITION BY n ORDER BY tuple()
        """, read_only=False)

        with self.assertRaises(CHException) as context:
            self.client.query_sync(f"""
                INSERT INTO {table_name} SELECT toDateTime('2021-01-01 00:00:00', 'UTC'), number as n FROM numbers(101)
            """, read_only=False)

        self.assertEqual(CHErrors.TOO_MANY_PARTS, context.exception.code)
        self.assertRegex(context.exception.args[0], r"\[Error\] Please make sure the ENGINE_PARTITION_KEY setting is correct. Large number of partitions is a common misconception. Partitioning is not intended to speed up SELECT queries \(the ENGINE_SORTING_KEY is sufficient to make range queries fast\), partitions are intended for data manipulation")

    def test_nested_aggregation_function(self):
        with self.assertRaises(CHException) as context:
            self.client.query_sync("""
                SELECT avg(number) AS number, max(number) FROM numbers(10)
            """)

        self.assertEqual(CHErrors.ILLEGAL_AGGREGATION, context.exception.code)
        self.assertEqual(context.exception.args[0], "[Error] You cannot use the aggregate function 'avg(number)' or its alias inside another aggregate function. Contact us at support@tinybird.co for help or read this SQL tip:  https://www.tinybird.co/docs/concepts/data-sources.html#partitioning")

    def test_nested_aggregation_function_with_alias(self):
        with self.assertRaises(CHException) as context:
            self.client.query_sync("""
                SELECT avg(max(number)) FROM numbers(10)
            """)

        self.assertEqual(CHErrors.ILLEGAL_AGGREGATION, context.exception.code)
        self.assertEqual(context.exception.args[0], "[Error] You cannot use the aggregate function 'max(number)' or its alias inside another aggregate function. Contact us at support@tinybird.co for help or read this SQL tip:  https://www.tinybird.co/docs/concepts/data-sources.html#partitioning")

    def test_get_columns_from_query(self):
        sync_ch_get_columns_from_query = async_to_sync(ch_get_columns_from_query)

        res = sync_ch_get_columns_from_query(self.database_host, self.test_database, "select 1 as example")

        self.assertEqual(res, [{
            'name': 'example',
            'normalized_name': 'example',
            'codec': None,
            'type': 'UInt8',
            'nullable': False,
            'auto': False,
            'default_value': None
        }])

    async def test_get_columns_from_complex_query(self):
        table_name = f"big_table_{uuid.uuid4().hex}"
        engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name + "','{replica}')"
        self.client.query_sync(f"""
            CREATE TABLE {table_name} ON CLUSTER tinybird
            ENGINE = {engine}
            ORDER BY number
            AS SELECT * FROM numbers_mt(10000000)
        """, read_only=False)

        start = time.monotonic()
        res = await ch_get_columns_from_query(
            self.database_host, self.test_database, f"""
            SELECT a.number AS first_col, b.number AS sec_col
            FROM {table_name} AS a CROSS JOIN (
                SELECT * from {table_name} PREWHERE number IN (
                    SELECT DISTINCT number FROM {table_name}
                    )
                ) AS b
            """)
        end = time.monotonic()

        self.assertLess(end - start, 0.5, "The operation needs to be fast")

        self.assertEqual(res, [
            {
                'auto': False,
                'codec': None,
                'default_value': None,
                'name': 'first_col',
                'normalized_name': 'first_col',
                'nullable': False,
                'type': 'UInt64'
            },
            {
                'auto': False,
                'codec': None,
                'default_value': None,
                'name': 'sec_col',
                'normalized_name': 'sec_col',
                'nullable': False,
                'type': 'UInt64'
            }
        ])


class TestUnicode(unittest.TestCase):

    def test_case_with_escaped_emojis_returns_ok(self):
        database_host = User.default_database_server
        ch_client = HTTPClient(database_host, database=None)
        headrs, body = ch_client.query_sync("select extractAll(JSONExtractString('{}', 'body'), '[\\\\x{1F600}-\\\\x{1F64F}|\\\\x{1F300}-\\\\x{1F5FF}|\\\\x{1F680}-\\\\x{1F6FF}|\\\\x{1F1E0}-\\\\x{1F1FF}|\\\\x{02702}-\\\\x{027B0}]') emojis FORMAT JSON", read_only=False)
        self.assertEqual([{"emojis": []}], json.loads(body)['data'])


class TestGlobalTimeout(unittest.IsolatedAsyncioTestCase):

    def tearDown(self) -> None:
        global_patch_for_requests_timeout(timeout=(60, 60))

    def test_global_read_timeout_is_applied(self):
        global_patch_for_requests_timeout(timeout=(0.05, 0.05))
        database_host = User.default_database_server
        ch_client = HTTPClient(database_host, database=None)
        with self.assertRaises(requests.exceptions.ReadTimeout):
            ch_client.query_sync("select sleepEachRow(1)", read_only=False)

    def test_global_read_timeout_is_applied_01(self):
        global_patch_for_requests_timeout(timeout=(0.05, 3))
        database_host = User.default_database_server
        ch_client = HTTPClient(database_host, database=None)
        headrs, body = ch_client.query_sync("select sleepEachRow(0.05) as a FORMAT JSON")
        self.assertEqual([{'a': 0}], json.loads(body)['data'])

    def test_global_read_timeout_can_be_overriden(self):
        global_patch_for_requests_timeout(timeout=(0.01, 0.01))
        database_host = User.default_database_server
        ch_client = HTTPClient(database_host, database=None)
        headrs, body = ch_client.query_sync("select sleepEachRow(0.05) as a FORMAT JSON", timeout=10)
        self.assertEqual([{'a': 0}], json.loads(body)['data'])

    async def test_global_read_timeout_is_reduced_by_query_max_execution_time_async(self):
        global_patch_for_requests_timeout(timeout=(10, 10))
        database_host = User.default_database_server
        ch_client = HTTPClient(database_host, database=None)

        query_id = str(uuid.uuid4())
        with self.assertRaises(CHException):
            start = time.time()
            _, _ = await ch_client.query("select sleepEachRow(1) a, sleepEachRow(0.001) as b from system.numbers LIMIT 3",
                                         query_id=query_id, max_execution_time=1, read_only=True)
            end = time.time()
            self.assertLess(end - start, 2)

        # Check that the query was killed
        # We don't check the exact query because the error changes between CH versions
        # and we need to wait for the query to finish (kill doesn't wake it up)
        _, _ = ch_client.query_sync("SYSTEM FLUSH LOGS", read_only=False)
        headers, body = ch_client.query_sync(f"SELECT count() as c FROM system.query_log "
                                             f"WHERE query iLIKE '%KILL%{query_id}%' "
                                             f"FORMAT JSON")
        res = json.loads(body)
        self.assertEqual(res['rows'], 1)
        self.assertTrue(res['data'][0]['c'], 1)


class TestCHTableQuerySubprocess(unittest.TestCase):

    def test_subprocess_timeout(self):
        columns = [
            {
                'name': 'c0',
                'normalized_name': 'c0',
                'type': 'String',
                'nullable': False,
                'auto': False
            },
            {
                'name': 'c1',
                'normalized_name': 'c1',
                'type': 'Int32',
                'nullable': False,
                'auto': False
            }
        ]
        chunk = """a,1\nb,2\nc,3""".encode()
        fast_query = 'select 0 as s, c0, c1 from table'
        slow_query = 'select sleepEachRow(1) as s, c0, c1 from table'  # simulates stuck subprocess
        t = CHTable(columns)

        output = t.query(chunk, fast_query, timeout=1)
        self.assertEqual(len(output['data']), 3)

        with self.assertRaisesRegex(Exception, r"CH return code: -9"):
            t.query(chunk, slow_query, timeout=0.1)


class TestExtractFromCSVExtract(unittest.TestCase):
    def test_custom_escapechar_easy(self):
        data = """123,"this is \\"just\\" one column, 3 in total",1234\n"""
        info = CSVInfo.extract_from_csv_extract(data, dialect_overrides={'escapechar': '\\', 'delimiter': ',', 'newline': None})
        self.assertEqual(len(info.columns), 3)

    def test_custom_escapechar(self):
        data = """1411820617817640960,0,0,0,0,2021-07-04 22:54:00,"{\\"authorId\\":\\"1475157055\\",\\"handle\\":\\"JustARedFlower\\",\\"userName\\":\\"everyone's mommy\\",\\"followers\\":540,\\"friends\\":774}","",RT @dirtyfourloko: It’s the 4th of July https://t.co/zlCgQBVZOI,"[{\\"type\\":\\"photo\\",\\"expanded_url\\":\\"https://twitter.com/dirtyfourloko/status/1411758352590323712/photo/1\\"}]",[],dirtyfourloko"""
        info = CSVInfo.extract_from_csv_extract(data, dialect_overrides={'escapechar': '\\'})
        self.assertEqual(len(info.columns), 12)


class TestTableDependentViews(unittest.TestCase):

    database_host = User.default_database_server

    test_database_1 = f'test_table_dependent_view_1_{uuid.uuid4().hex}'
    test_database_2 = f'test_table_dependent_view_2_{uuid.uuid4().hex}'

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {cls.test_database_1} ON CLUSTER tinybird SYNC", read_only=False)
        ch_client_for_setup.query_sync(f"CREATE DATABASE {cls.test_database_1} ON CLUSTER tinybird", read_only=False)
        ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {cls.test_database_2} ON CLUSTER tinybird SYNC", read_only=False)
        ch_client_for_setup.query_sync(f"CREATE DATABASE {cls.test_database_2} ON CLUSTER tinybird", read_only=False)
        """
        db test_table_dependent_view_1:
        - ds1
        - mv1: ds1 → ds2
        - ds2
        - table_without_deps

        db test_table_dependent_view_2:
        - mv2: testdb1.ds1 → ds3
        - ds3
        """

        engine_1 = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + f"{cls.test_database_1}.ds1" + "','{replica}')"
        engine_2 = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + f"{cls.test_database_1}.table_without_deps" + "','{replica}')"
        engine_summing_1 = "ReplicatedSummingMergeTree('/clickhouse/tables/{layer}-{shard}/" + f"{cls.test_database_1}.ds2" + "','{replica}')"
        engine_summing_2 = "ReplicatedSummingMergeTree('/clickhouse/tables/{layer}-{shard}/" + f"{cls.test_database_2}.ds3" + "','{replica}')"
        ch_client_for_setup.query_sync(f"CREATE TABLE {cls.test_database_1}.ds1 ON CLUSTER tinybird (value Float32) ENGINE={engine_1} ORDER BY tuple()", read_only=False)
        ch_client_for_setup.query_sync(f"CREATE TABLE {cls.test_database_1}.ds2 ON CLUSTER tinybird (count UInt64) ENGINE = {engine_summing_1} PARTITION BY tuple() ORDER BY tuple()", read_only=False)
        ch_client_for_setup.query_sync(f"CREATE MATERIALIZED VIEW {cls.test_database_1}.mv1 ON CLUSTER tinybird TO {cls.test_database_1}.ds2 AS SELECT count(*) as count FROM {cls.test_database_1}.ds1", read_only=False)
        ch_client_for_setup.query_sync(f"CREATE TABLE {cls.test_database_1}.table_without_deps ON CLUSTER tinybird (value Float32) ENGINE={engine_2} ORDER BY tuple();", read_only=False)

        ch_client_for_setup.query_sync(f"CREATE TABLE {cls.test_database_2}.ds3 ON CLUSTER tinybird (count UInt64) ENGINE = {engine_summing_2} PARTITION BY tuple() ORDER BY tuple();", read_only=False)
        ch_client_for_setup.query_sync(f"CREATE MATERIALIZED VIEW {cls.test_database_2}.mv2 ON CLUSTER tinybird TO {cls.test_database_2}.ds3 AS SELECT count(*) as count FROM {cls.test_database_1}.ds1", read_only=False)

    @classmethod
    def tearDownClass(cls):
        ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {cls.test_database_1} ON CLUSTER tinybird SYNC", read_only=False)
        ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {cls.test_database_2} ON CLUSTER tinybird SYNC", read_only=False)
        super().tearDownClass()

    def test_dependent_materialized_views_returned_even_when_they_are_in_different_databases(self):
        res = ch_table_dependent_views_sync(self.database_host, self.test_database_1, 'ds1')
        self.assertEqual(2, len(res))
        self.assertIn(CHTableLocation(self.test_database_1, 'mv1'), res)
        self.assertIn(CHTableLocation(self.test_database_2, 'mv2'), res)

    def test_table_without_dependent_views_return_an_empty_result(self):
        res = ch_table_dependent_views_sync(self.database_host, self.test_database_1, 'table_without_deps')
        self.assertEqual([], res)


class TestXClickHouseSummaryHeader(unittest.IsolatedAsyncioTestCase):

    def setUp(self):
        super().setUp()
        self.database_host = User.default_database_server
        rand_id = uuid.uuid4().hex
        self.test_database = f'x_ch_summary_{rand_id}'
        self.ch_client_for_setup = HTTPClient(self.database_host, database=None)
        self.ch_client_for_setup.query_sync(f"CREATE DATABASE IF NOT EXISTS {self.test_database} ON CLUSTER tinybird", read_only=False)

        self.client = HTTPClient(self.database_host, database=self.test_database)

    def tearDown(self):
        self.ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {self.test_database} ON CLUSTER tinybird SYNC", read_only=False)
        super().tearDown()

    def test_x_clickhouse_summary_header_insert(self):
        CH_VERSION = get_clickhouse_version()

        table_name = 'summary_table'
        self.client.query_sync(f"CREATE TABLE {table_name} ON CLUSTER tinybird (a String) Engine=Null()", read_only=False)
        headers, _ = self.client.insert_chunk(f'INSERT INTO {table_name} FORMAT CSV', 'value')
        obj = json.loads(headers['X-ClickHouse-Summary'])

        expected_read_rows = '1' if CH_VERSION >= 22.6 else '0'
        expected_read_bytes = '14' if CH_VERSION >= 22.6 else '0'

        self.assertEqual(obj['read_rows'], expected_read_rows, obj)
        self.assertEqual(obj['read_bytes'], expected_read_bytes, obj)
        self.assertEqual(obj['written_rows'], '1', obj)
        self.assertEqual(obj['written_bytes'], '14', obj)
        self.assertEqual(obj['total_rows_to_read'], '0', obj)

        if CH_VERSION >= 22.8:
            self.assertTrue(obj['result_rows'] == '1', obj)
            self.assertTrue(obj['result_bytes'] == '14', obj)

    def test_x_clickhouse_summary_header_insert_with_mv(self):
        table_name = "floats"
        engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name + "','{replica}')"
        self.client.query_sync(f"CREATE TABLE IF NOT EXISTS {table_name} ON CLUSTER tinybird (v Float64) Engine={engine} order by tuple()", read_only=False)

        headers, _ = self.client.insert_chunk('INSERT INTO floats FORMAT CSV', csv_from_python_object([[1.0]]))

        obj = json.loads(headers['X-ClickHouse-Summary'])

        self.assertIn(obj['read_rows'], ['0', '1'], obj)  # 0 in old CH releases, 1 in 22.6+
        self.assertIn(obj['read_bytes'], ['0', '8'], obj)  # 0 in old CH releases, 8 in 22.6+
        self.assertEqual(obj['written_rows'], '1', obj)
        self.assertEqual(obj['written_bytes'], '8', obj)
        self.assertEqual(obj['total_rows_to_read'], '0', obj)
        self.assertTrue('result_rows' not in obj or obj['result_rows'] == '1', obj)  # Added in 22.8+
        self.assertTrue('result_bytes' not in obj or obj['result_bytes'] == '8', obj)  # Added in 22.8+

        table_name_target = "target"
        engine_target = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name_target + "','{replica}')"
        table_name_target2 = "target2"
        engine_target2 = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name_target2 + "','{replica}')"
        self.client.query_sync(f"CREATE TABLE {table_name_target} ON CLUSTER tinybird (v Float64) Engine={engine_target} order by tuple()", read_only=False)
        self.client.query_sync(f"CREATE TABLE {table_name_target2} ON CLUSTER tinybird (v Float64) Engine={engine_target2} order by tuple()", read_only=False)
        self.client.query_sync(f"CREATE MATERIALIZED VIEW {self.test_database}.floats_to_target ON CLUSTER tinybird TO {self.test_database}.target AS SELECT * FROM {self.test_database}.floats",
                               read_only=False)
        self.client.query_sync(f"CREATE MATERIALIZED VIEW {self.test_database}.floats_to_target2 ON CLUSTER tinybird TO {self.test_database}.target2 AS SELECT * FROM {self.test_database}.floats",
                               read_only=False)

        # Replicated deduplication prevent us from ingesting twice the same value here. Using a different float is key here.
        # See https://clickhouse.com/docs/en/operations/settings/merge-tree-settings/#replicated-deduplication-window
        headers, _ = self.client.insert_chunk('INSERT INTO floats FORMAT CSV', csv_from_python_object([[2.0]]))

        obj = json.loads(headers['X-ClickHouse-Summary'])

        # [rows, bytes]
        prev_22_6 = ['2', '16']
        post_22_6 = ['1', '8']  # MV not included
        from_22_9 = ['3', '24']  # Landing included
        self.assertIn(obj['read_rows'], [prev_22_6[0], post_22_6[0], from_22_9[0]], obj)
        self.assertIn(obj['read_bytes'], [prev_22_6[1], post_22_6[1], from_22_9[1]], obj)
        self.assertEqual(obj['written_rows'], '3', obj)
        self.assertEqual(obj['written_bytes'], '24', obj)
        self.assertEqual(obj['total_rows_to_read'], '0', obj)
        self.assertTrue('result_rows' not in obj or obj['result_rows'] == '3', obj)  # Added in 22.8+
        self.assertTrue('result_bytes' not in obj or obj['result_bytes'] == '24', obj)  # Added in 22.8+

    async def test_x_clickhouse_summary_select(self):
        table_name = "floats"
        engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name + "','{replica}')"
        self.client.query_sync(f"CREATE TABLE {table_name} ON CLUSTER tinybird (v Float64) Engine={engine} order by tuple()", read_only=False)
        headers, _ = self.client.insert_chunk('INSERT INTO floats FORMAT CSV', csv_from_python_object([[1.0]]))

        query_settings = {
            # To get X-ClickHouse-Summary filled we need CH 22.6 or setting send_progress_in_http_header
            # https://github.com/ClickHouse/ClickHouse/issues/16207
            'send_progress_in_http_headers': 1,
            # Avoid X-ClickHouse-Progress setting big interval
            # https://github.com/ClickHouse/ClickHouse/issues/12122
            # https://github.com/ClickHouse/ClickHouse/issues/33630
            'http_headers_progress_interval_ms': 100000,
        }

        headers, _ = await self.client.query('SELECT * FROM floats', **query_settings)

        obj = json.loads(headers['X-ClickHouse-Summary'])

        self.assertEqual(obj['read_rows'], '1', obj)
        self.assertEqual(obj['read_bytes'], '8', obj)
        self.assertEqual(obj['written_rows'], '0', obj)
        self.assertEqual(obj['written_bytes'], '0', obj)
        self.assertEqual(obj['total_rows_to_read'], '1', obj)
        self.assertTrue('result_rows' not in obj or obj['result_rows'] == '1', obj)  # Added in 22.8+
        self.assertTrue('result_bytes' not in obj or  # noqa: W504
                        # Added in 22.8
                        # 22.10 added more padding to internal structures which means more memory usage
                        # https://github.com/ClickHouse/ClickHouse/pull/42564
                        # Minimum padding before: 64. 22.10+: 256
                        obj['result_bytes'] in ['64', '256'], obj)


class TestOnCluster(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.database_host = User.default_database_server
        cls.test_database = f'd_oncluster_{uuid.uuid4().hex}'

        ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        ch_client_for_setup.query_sync(f"CREATE DATABASE IF NOT EXISTS {cls.test_database} ON CLUSTER tinybird", read_only=False)

        cls.client = HTTPClient(cls.database_host, database=cls.test_database)

    @classmethod
    def tearDownClass(cls):
        ch_client_for_setup = HTTPClient(cls.database_host, database=None)
        ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {cls.test_database} ON CLUSTER tinybird SYNC", read_only=False)
        super().tearDownClass()

    async def test_get_query_log_on_cluster(self):
        with patch.object(HTTPClient, 'query', autospec=True) as mock_query_log:
            await _get_query_log(self.database_host, self.test_database, 'query_id', cluster='tinybird')
            query = f"\n        SELECT\n            type,\n            event_time,\n            exception\n        FROM clusterAllReplicas('tinybird', system.query_log)\n        WHERE\n            event_time > now() - INTERVAL 2 DAY\n            AND current_database = '{self.test_database}'\n            AND query_id = 'query_id'\n        ORDER BY\n            event_time DESC\n        FORMAT JSON\n    "
            mock_query_log.assert_called_once_with(ANY, query, max_execution_time=2)

    async def test_get_query_log(self):
        with patch.object(HTTPClient, 'query', autospec=True) as mock_query_log:
            await _get_query_log(self.database_host, self.test_database, 'query_id')
            query = f"\n        SELECT\n            type,\n            event_time,\n            exception\n        FROM system.query_log\n        WHERE\n            event_time > now() - INTERVAL 2 DAY\n            AND current_database = '{self.test_database}'\n            AND query_id = 'query_id'\n        ORDER BY\n            event_time DESC\n        FORMAT JSON\n    "
            mock_query_log.assert_called_once_with(ANY, query, max_execution_time=2)

    async def test_ch_flush_logs_on_all_replicas(self):
        clusters = await ch_get_cluster_instances(self.database_host, self.test_database, 'tinybird')

        with patch('tinybird.ch.ch_get_cluster_instances', return_value=clusters):
            with patch.object(HTTPClient, 'query', autospec=True) as mock_query_log:
                await ch_flush_logs_on_all_replicas(self.database_host, 'tinybird', user_agent='test')
                self.assertEqual(mock_query_log.call_count, 2)
                self.assertEqual(mock_query_log.call_args_list[0][0][1], "SYSTEM FLUSH LOGS")
                self.assertEqual(mock_query_log.call_args_list[0][1], {'read_only': False, 'user_agent': 'test', 'max_execution_time': 6})

    def test_ch_flush_logs_on_all_replicas_sync(self):
        clusters = async_to_sync(ch_get_cluster_instances)(self.database_host, self.test_database, 'tinybird')

        with patch('tinybird.ch.ch_get_cluster_instances', return_value=clusters):
            with patch.object(HTTPClient, 'query', autospec=True) as mock_query_log:
                ch_flush_logs_on_all_replicas_sync(self.database_host, 'tinybird', user_agent='test')
                self.assertEqual(mock_query_log.call_count, 2)
                self.assertEqual(mock_query_log.call_args_list[0][0][1], "SYSTEM FLUSH LOGS")
                self.assertEqual(mock_query_log.call_args_list[0][1], {'read_only': False, 'user_agent': 'test', 'max_execution_time': 6})


class TestTruncateTable(unittest.IsolatedAsyncioTestCase):
    database_server = User.default_database_server
    rand_id = uuid.uuid4().hex
    database = f'database_name_{rand_id}'
    table = f'table_name_{rand_id}'
    new_table = f'new_table_name_{rand_id}'
    cluster = User.default_cluster

    @classmethod
    def setUpClass(cls):
        engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + f"{cls.database}.{cls.table}" + "','{replica}')"
        cls.ch_client_for_setup = HTTPClient(cls.database_server, database=None)
        cls.ch_client_for_setup.query_sync(f"CREATE DATABASE {cls.database} ON CLUSTER {cls.cluster}", read_only=False)
        # We create a special table with has 2 settings things
        # A storage_policy which we could have introduced via feature flag
        # A setting we might have added manually to CH
        cls.ch_client_for_setup.query_sync(f"""
        CREATE TABLE {cls.database}.{cls.table} ON CLUSTER {cls.cluster}
        (
            VisitDate Date,
            Hour UInt8
        )
        ENGINE = {engine}
        PARTITION BY toYYYYMM(VisitDate)
        ORDER BY Hour
        SETTINGS storage_policy = '{OTHER_STORAGE_POLICY}', max_parts_in_total=50000
        """, read_only=False)

    def tearDown(self):
        self.ch_client_for_setup.query_sync(f"DROP DATABASE IF EXISTS {self.database} ON CLUSTER {self.cluster} SYNC", read_only=False)
        super().tearDown()

    async def test_ch_create_exact_empty_copy(self):
        await ch_create_exact_empty_copy(self.database_server, self.database,
                                         self.table, self.cluster,
                                         self.new_table)

        table_details = await ch_table_details_async(self.table, self.database_server, self.database)
        new_table_details = await ch_table_details_async(self.new_table, self.database_server, self.database)
        table_meta = table_details.to_json(exclude=['engine_full'])
        new_table_meta = new_table_details.to_json(exclude=['engine_full'])
        self.assertEquals(table_meta, new_table_meta)
        self.assertTrue(table_details.is_replicated())
        self.assertTrue(new_table_details.is_replicated())

        # Make sure both settings survived the copy
        table_details_full = await ch_table_details_async(self.table, self.database_server, self.database, clean_settings=False)
        new_table_details_full = await ch_table_details_async(self.new_table, self.database_server, self.database, clean_settings=False)
        table_meta_full = table_details_full.to_json(exclude=['engine_full'])
        new_table_meta_full = new_table_details_full.to_json(exclude=['engine_full'])
        self.assertEquals(table_meta_full, new_table_meta_full)
        self.assertIn('storage_policy', new_table_details_full.engine_full)
        self.assertIn('storage_policy', new_table_details_full.settings)
        self.assertIn('max_parts_in_total', new_table_details_full.engine_full)
        self.assertIn('max_parts_in_total', new_table_details_full.settings)

    async def test_ch_truncate_table(self):
        table_to_truncate = f"table_to_truncate_{self.rand_id}"
        engine = "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_to_truncate + "','{replica}')"
        ch_client = HTTPClient(self.database_server, database=None)
        ch_client.query_sync(f"DROP DATABASE IF EXISTS {self.database} ON CLUSTER {self.cluster} SYNC", read_only=False)
        ch_client.query_sync(f"CREATE DATABASE {self.database} ON CLUSTER {self.cluster}", read_only=False)
        ch_client.query_sync(f"""
        CREATE TABLE {self.database}.{table_to_truncate} ON CLUSTER {self.cluster}
        (
            VisitDate Date,
            Hour UInt8
        )
        ENGINE = {engine}
        PARTITION BY toYYYYMM(VisitDate)
        ORDER BY Hour;
        """, read_only=False)
        ch_client.query_sync(f"""
        INSERT INTO {self.database}.{table_to_truncate} (*)
        VALUES ('2022-08-24',7)""", read_only=False)
        previous_response = ch_client.query_sync(f"""
        SELECT * from {self.database}.{table_to_truncate}""", read_only=False)[-1]
        self.assertTrue(previous_response.decode('utf-8'))
        await ch_truncate_table(self.database_server, self.database,
                                table_to_truncate, self.cluster)
        truncated_response = ch_client.query_sync(f"""
        SELECT * from {self.database}.{table_to_truncate}""", read_only=False)[-1]
        self.assertFalse(truncated_response.decode('utf-8'))

    async def test_ch_truncate_table_with_fallback_happy_path(self):
        with unittest.mock.patch('tinybird.ch.ch_truncate_table') as mock_ch_truncate_table:
            await ch_truncate_table_with_fallback(self.database_server, self.database, self.table, self.cluster)
            mock_ch_truncate_table.assert_called_once()

    async def test_ch_truncate_table_with_fallback_generic_error(self):
        ch_exception = CHException('Code: 32, e.displayText() = DB::Exception: Attempt to read after eof: Cannot parse Int16 from String, because value is too short (version 19.13.1.1)')
        with unittest.mock.patch('tinybird.ch.ch_truncate_table', side_effect=ch_exception):
            with self.assertRaises(CHException):
                await ch_truncate_table_with_fallback(self.database_server, self.database, self.table, self.cluster)

    async def test_ch_truncate_table_with_fallback_table_size_exceeds_max_drop_size(self):
        ch_exception = CHException('Code: 359, e.displayText = DB::Exception: Table or Partition in table_name was not dropped.')
        with unittest.mock.patch('tinybird.ch.ch_truncate_table', side_effect=ch_exception):
            with unittest.mock.patch('tinybird.ch.ch_create_exact_empty_copy') as mock_ch_create_exact_empty_copy:
                with unittest.mock.patch('tinybird.ch.ch_swap_tables') as mock_ch_swap_tables:
                    await ch_truncate_table_with_fallback(self.database_server, self.database, self.table, self.cluster)
                    mock_ch_create_exact_empty_copy.assert_called_once()
                    mock_ch_swap_tables.assert_called_once()


async def wait_for_query_start(client, query_id):
    i = 0
    while i < 10:
        query = f"""SELECT count() c from clusterAllReplicas('tinybird', 'system.processes') where query_id='{query_id}' format JSON"""
        _, body = await client.query(query)
        data = json.loads(body)['data']
        if data[0]['c'] >= 1:
            break
        try:
            await asyncio.sleep(0.05)
        except RuntimeError:
            time.sleep(0.05)
        i += 1
    if i == 10:
        raise RuntimeError('query did not start')


class TestQuery(unittest.IsolatedAsyncioTestCase):

    async def test_response_is_compress(self):
        database_host = User.default_database_server
        client = HTTPClient(database_host, database=None)

        headers, body = await client.query("SELECT number FROM numbers(10) FORMAT JSON", compress=True)
        body = zlib.decompress(body, wbits=16 + 15)

        rows = json.loads(body).get('data', [])
        self.assertEqual(len(rows), 10, body)
        self.assertEqual(headers['Content-Encoding'], 'gzip', headers)
